# cython: boundscheck=False, embedsignature=True

"""Coordinate reference systems, the CRS class and supporting functions.

A coordinate reference system (CRS) defines how a dataset's pixels map
to locations on, for example, a globe or the Earth. A CRS may be local
or global. The GIS field shares a number of authority files that define
CRS. "EPSG:32618" is the name of a regional CRS from the European
Petroleum Survey Group authority file. "OGC:CRS84" is the name of a
global CRS from the Open Geospatial Consortium authority. Custom CRS can
be described in text using several formats. Rasterio's CRS class is our
abstraction for coordinate reference systems.

A rasterio dataset's crs property is an instance of CRS. CRS are also
used to define transformations between coordinate reference systems.
These transformations are performed by the PROJ library. Rasterio does
not call PROJ functions directly, but invokes them via calls to GDAL's
"OSR*" functions.

"""

from collections import defaultdict
from itertools import groupby
import json
import logging
from operator import itemgetter
import pickle
import typing
import warnings
import re

import rasterio._env
from rasterio._err import CPLE_BaseError, CPLE_NotSupportedError
from rasterio.errors import CRSError, RasterioDeprecationWarning
from rasterio.enums import WktVersion

from rasterio._base cimport osr_set_traditional_axis_mapping_strategy
from rasterio._err cimport exc_wrap_ogrerr, exc_wrap_int, exc_wrap_pointer


log = logging.getLogger(__name__)


_RE_PROJ_PARAM = re.compile(r"""
    \+              # parameter starts with '+' character
    (?P<param>\w+)    # capture parameter name
    \=?             # match both key only and key-value parameters
    (?P<value>\S+)? # capture all characters up to next space (None if no value)
    \s*?            # consume remaining whitespace, if any
""", re.X)


def auth_preference(item):
    preferred_order = ["EPSG", "OGC", "ESRI", "IAU_2015"]
    conf, auth, val = item
    if auth in preferred_order:
        return preferred_order.index(auth)
    else:
        return 100


cdef _safe_osr_release(OGRSpatialReferenceH srs):
    """Wrapper to handle OSR release when NULL."""
    if srs != NULL:
        OSRRelease(srs)
    srs = NULL


cdef class CRS:
    """A geographic or projected coordinate reference system.

    CRS objects may be created by passing PROJ parameters as keyword
    arguments to the standard constructor or by passing EPSG codes, PROJ
    mappings, PROJ strings, or WKT strings to the from_epsg, from_dict,
    from_string, or from_wkt static methods.

    Examples
    --------

    The from_dict method takes PROJ parameters as keyword arguments.

    >>> crs = CRS.from_dict(proj="aea")

    EPSG codes may be used with the from_epsg method.

    >>> crs = CRS.from_epsg(3005)

    The from_string method takes a variety of input.

    >>> crs = CRS.from_string("EPSG:3005")
    """
    def __init__(self, initialdata=None, **kwargs):
        """Make a CRS from a PROJ dict or mapping.

        Parameters
        ----------
        initialdata : mapping, optional
            A dictionary or other mapping
        kwargs : mapping, optional
            Another mapping. Will be overlaid on the initialdata.

        Returns
        -------
        CRS

        """
        cdef CRS tmp
        self._data = {}
        self._epsg = None
        self._wkt = None
        self._geodetic_crs = None

        if initialdata or kwargs:
            tmp = CRS.from_dict(initialdata=initialdata, **kwargs)
            self._osr = OSRClone(tmp._osr)
            self._wkt = tmp._wkt
            self._data = tmp.data
            self._epsg = tmp._epsg

    @property
    def data(self):
        """A PROJ4 dict representation of the CRS.
        """
        if not self._data:
            self._data = self.to_dict()
        return self._data

    @property
    def is_valid(self):
        """Test that the CRS is a geographic or projected CRS.

        .. deprecated:: 1.4.0
           This property is not useful and will be removed in 2.0.0.

        Returns
        -------
        bool

        """
        warnings.warn("is_valid is not useful and will be removed in 2.0.0.", RasterioDeprecationWarning)
        return self.is_geographic or self.is_projected

    @property
    def is_epsg_code(self):
        """Test if the CRS is defined by an EPSG code.

        Returns
        -------
        bool

        """
        try:
            return bool(self.to_epsg())
        except CRSError:
            return False

    @property
    def wkt(self):
        """An OGC WKT representation of the CRS

        Returns
        -------
        str

        """
        if not self._wkt:
            self._wkt = self.to_wkt()
        return self._wkt

    @property
    def is_geographic(self):
        """Test if the CRS is a geographic coordinate reference system.

        Returns
        -------
        bool

        Raises
        ------
        CRSError

        """
        try:
            return bool(OSRIsGeographic(self._osr) == 1)
        except CPLE_BaseError as exc:
            raise CRSError("{}".format(exc))

    @property
    def is_projected(self):
        """Test if the CRS is a projected coordinate reference system.

        Returns
        -------
        bool

        Raises
        ------
        CRSError

        """
        try:
            return bool(OSRIsProjected(self._osr) == 1)
        except CPLE_BaseError as exc:
            raise CRSError("{}".format(exc))

    @property
    def linear_units(self):
        """Get a short name for the linear units of the CRS.

        Returns
        -------
        units : str
            "m", "ft", etc.

        Raises
        ------
        CRSError

        """
        try:
            return self.linear_units_factor[0]
        except CRSError:
            return "unknown"

    @property
    def linear_units_factor(self):
        """Get linear units and the conversion factor to meters of the CRS.

        Returns
        -------
        units : str
            "m", "ft", etc.
        factor : float
            Ratio of one unit to one meter.    

        Raises
        ------
        CRSError

        """
        cdef char *units_c = NULL
        cdef double to_meters

        try:
            if self.is_projected:
                to_meters = OSRGetLinearUnits(self._osr, &units_c)
            else:
                raise CRSError("Linear units factor is not defined for non projected CRS")
        except CPLE_BaseError as exc:
            raise CRSError("{}".format(exc))
        else:
            units_b = units_c
            return (units_b.decode('utf-8'), to_meters)

    @property
    def units_factor(self):
        """Get units and the conversion factor of the CRS.

        Returns
        -------
        units : str
            "m", "ft", etc.
        factor : float
            Ratio of one unit to one radian if the CRS is geographic
            otherwise, it is to one meter.

        Raises
        ------
        CRSError

        """
        cdef char *units_c = NULL
        cdef double factor

        try:
            if self.is_geographic:
                factor = OSRGetAngularUnits(self._osr, &units_c)
            else:
                factor = OSRGetLinearUnits(self._osr, &units_c)
        except CPLE_BaseError as exc:
            raise CRSError(exc)
        else:
            units_b = units_c
            return (units_b.decode('utf-8'), factor)

    @property
    def geodetic_crs(self):
        """Get the Geographic CRS from the CRS.

        Returns
        -------
        CRS

        Raises
        ------
        CRSError

        """
        if self._geodetic_crs:
            return self._geodetic_crs
        cdef CRS obj = CRS.__new__(CRS)
        try:
            obj._osr = exc_wrap_pointer(OSRCloneGeogCS(self._osr))
        except CPLE_BaseError as exc:
            raise CRSError("Cannot determine Geodetic CRS. {}".format(exc))
        else:
            osr_set_traditional_axis_mapping_strategy(obj._osr)
            self._geodetic_crs = obj
        return self._geodetic_crs 
        

    def to_dict(self, projjson=False):
        """Convert CRS to a PROJ dict.

        .. note:: If there is a corresponding EPSG code, it will be used
           when returning PROJ parameter dict.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        projjson: bool, default=False
            If True, will convert to PROJ JSON dict (Requites GDAL 3.1+
            and PROJ 6.2+).  If False, will convert to PROJ parameter
            dict.

        Returns
        -------
        dict

        """
        cdef OGRSpatialReferenceH osr = NULL
        cdef char *proj_c = NULL

        if projjson:
            text = self._projjson()
            return json.loads(text) if text else {}

        else:
            try:
                osr = exc_wrap_pointer(OSRClone(self._osr))
                exc_wrap_ogrerr(OSRExportToProj4(osr, &proj_c))

            except CPLE_BaseError as exc:
                return {}
                # raise CRSError("The WKT could not be parsed. {}".format(exc))

            else:
                proj_b = proj_c
                proj = proj_b.decode('utf-8')

            finally:
                CPLFree(proj_c)
                _safe_osr_release(osr)

            def parse(v):
                try:
                    return int(v)
                except ValueError:
                    pass
                try:
                    return float(v)
                except ValueError:
                    return v

            rv = {}
            for param in _RE_PROJ_PARAM.finditer(proj):
                key, value = param.groups()
                if key not in all_proj_keys:
                    continue

                if value is None or value.lower() == "true":
                    rv[key] = True
                elif value.lower() == "false":
                    continue
                else:
                    rv[key] = parse(value)
            return rv

    def to_proj4(self):
        """Convert to a PROJ4 representation.

        Returns
        -------
        str

        """
        return ' '.join(['+{}={}'.format(key, val) for key, val in self.data.items()])

    def to_wkt(self, morph_to_esri_dialect=False, version=None):
        """Convert to a OGC WKT representation.

         .. versionadded:: 1.3.0 version

        Parameters
        ----------
        morph_to_esri_dialect : bool, optional
            Whether or not to morph to the Esri dialect of WKT Only
            applies to GDAL versions < 3. This parameter will be removed
            in a future version of rasterio.
        version : WktVersion or str, optional
            The version of the WKT output.
            Only works with GDAL 3+. Default is WKT1_GDAL.

        Returns
        -------
        str

        Raises
        ------
        CRSError

        """
        cdef char *conv_wkt = NULL
        cdef const char* options_wkt[2]
        options_wkt[0] = NULL
        options_wkt[1] = NULL

        try:
            if OSRGetName(self._osr) != NULL:
                if morph_to_esri_dialect:
                    warnings.warn(
                        "'morph_to_esri_dialect' ignored with GDAL 3+. "
                        "Use 'version=WktVersion.WKT1_ESRI' instead."
                    )
                if version:
                    version = WktVersion(version).value
                    wkt_format = "FORMAT={}".format(version).encode("utf-8")
                    options_wkt[0] = wkt_format
                exc_wrap_ogrerr(OSRExportToWktEx(self._osr, &conv_wkt, options_wkt))
        except CPLE_BaseError as exc:
            raise CRSError("Cannot convert to WKT. {}".format(exc))

        else:
            if conv_wkt != NULL:
                return conv_wkt.decode('utf-8')
            else:
                return ''
        finally:
            CPLFree(conv_wkt)


    def to_epsg(self, confidence_threshold=70):
        """Convert to the best match EPSG code.

        For a CRS created using an EPSG code, that same value is
        returned.  For other CRS, including custom CRS, an attempt is
        made to match it to definitions in the EPSG authority file.
        Matches with a confidence below the threshold are discarded.

        Parameters
        ----------
        confidence_threshold : int
            Percent match confidence threshold (0-100).
           
        Returns
        -------
        int or None

        Raises
        ------
        CRSError

        """
        if self._epsg is not None:
            return self._epsg
        else:
            matches = self._matches(confidence_threshold=confidence_threshold)
            if not matches:
                return None
            else:
                for conf, auth, val in matches:
                    if auth == "EPSG":
                        return int(val)
        return None

    def to_authority(self, confidence_threshold=70):
        """Convert to the best match authority name and code.

        For a CRS created using an EPSG code, that same value is
        returned.  For other CRS, including custom CRS, an attempt is
        made to match it to definitions in authority files.  Matches
        with a confidence below the threshold are discarded.

        Parameters
        ----------
        confidence_threshold : int
            Percent match confidence threshold (0-100).

        Returns
        -------
        name : str
            Authority name.
        code : str
            Code from the authority file.

        or None

        """
        if self._epsg is not None:
            return ("EPSG", str(self._epsg))
        else:
            matches = self._matches(confidence_threshold=confidence_threshold)
            if not matches:
                return None

            grouped = groupby(matches, key=itemgetter(0))
            for k, group in grouped:
                conf, auth, val = sorted(group, key=auth_preference)[0]
                return auth, str(val)

    def _matches(self, confidence_threshold=70):
        """Find matches in authority files.

        Returns
        -------
        dict : {name: [codes]}
            A dictionary in which capitalized authority names are the
            keys and lists of codes ordered by match confidence,
            descending, are the values.

        """
        cdef OGRSpatialReferenceH osr = NULL
        cdef OGRSpatialReferenceH *matches = NULL
        cdef int *confidences = NULL
        cdef int num_matches = 0
        cdef int i = 0
        cdef list results = []

        try:
            osr = exc_wrap_pointer(OSRClone(self._osr))
            matches = OSRFindMatches(osr, NULL, &num_matches, &confidences)

            for i in range(num_matches):
                confidence = confidences[i]
                c_code = OSRGetAuthorityCode(matches[i], NULL)
                c_name = OSRGetAuthorityName(matches[i], NULL)

                if c_code == NULL:
                    log.debug("returned authority code was null")
                if c_name == NULL:
                    log.debug("returned authority name was null")

                if c_code != NULL and c_name != NULL and confidence >= confidence_threshold:
                    code = c_code.decode('utf-8')
                    name = c_name.decode('utf-8')
                    results.append((confidence, name, code))

            return results

        finally:
            _safe_osr_release(osr)
            OSRFreeSRSArray(matches)
            CPLFree(confidences)

    def equals(self, other, ignore_axis_order=False):
        """
        Check if the crs objects are equivalent.

        Properties
        ----------
        other: CRS
            the other CRS to compare to
        ignore_axis_order: bool, default=False
            If True, it will compare the CRS class and ignore the axis order.

        Returns
        -------
        bool
        """
        cdef CRS crs_o
        cdef const char* options[2]
        
        if ignore_axis_order:
            options[0] = b"IGNORE_DATA_AXIS_TO_SRS_AXIS_MAPPING=YES"
        else:
            options[0] = b"IGNORE_DATA_AXIS_TO_SRS_AXIS_MAPPING=NO"
        options[1] = NULL

        try:
            crs_o = CRS.from_user_input(other)
            return bool(OSRIsSameEx(self._osr, crs_o._osr, options) == 1)
        except CRSError:
            return False
        

    def to_string(self):
        """Convert to a PROJ4 or WKT string.

        The output will be reduced as much as possible by attempting a
        match to CRS defined in authority files.

        Notes
        -----
        Mapping keys are tested against the ``all_proj_keys`` list.
        Values of ``True`` are omitted, leaving the key bare:
        {'no_defs': True} -> "+no_defs" and items where the value is
        otherwise not a str, int, or float are omitted.

        Returns
        -------
        str

        Raises
        ------
        CRSError

        """
        auth = self.to_authority()
        if auth:
            return ":".join(auth)
        else:
            return self.to_wkt() or self.to_proj4()

    @staticmethod
    def from_epsg(code):
        """Make a CRS from an EPSG code.

        Parameters
        ----------
        code : int or str
            An EPSG code. Strings will be converted to integers.

        Notes
        -----
        The input code is not validated against an EPSG database.

        Returns
        -------
        CRS

        Raises
        ------
        CRSError

        """
        cdef CRS obj = CRS.__new__(CRS)

        try:
            code = int(code)
        except OverflowError as err:
            raise CRSError(f"Not in the range of valid EPSG codes: {code}") from err
        except TypeError as err:
            raise CRSError(f"Not a valid EPSG codes: {code}") from err

        if code <= 0:
            raise CRSError("EPSG codes are positive integers")

        try:
            exc_wrap_ogrerr(exc_wrap_int(OSRImportFromEPSG(obj._osr, <int>code)))
        except OverflowError as err:
            raise CRSError(f"Not in the range of valid EPSG codes: {code}") from err
        except CPLE_BaseError as exc:
            raise CRSError("The EPSG code is unknown. {}".format(exc))
        else:
            osr_set_traditional_axis_mapping_strategy(obj._osr)
            obj._epsg = code
            return obj

    @staticmethod
    def from_proj4(proj):
        """Make a CRS from a PROJ4 string.

        Parameters
        ----------
        proj : str
            A PROJ4 string like "+proj=longlat ..."

        Returns
        -------
        CRS

        Raises
        ------
        CRSError

        """
        cdef CRS obj = CRS.__new__(CRS)

        # Filter out nonsensical items that might have crept in.
        items_filtered = []
        for param in _RE_PROJ_PARAM.finditer(proj):
            value = param.group('value')
            if value is None:
                items_filtered.append(param.group())
            elif value.lower() == "false":
                continue
            else:
                items_filtered.append(param.group())
                
        proj = ' '.join(items_filtered)
        proj_b = proj.encode('utf-8')

        try:
            exc_wrap_ogrerr(exc_wrap_int(OSRImportFromProj4(obj._osr, <const char *>proj_b)))
        except CPLE_BaseError as exc:
            raise CRSError("The PROJ4 dict could not be understood. {}".format(exc))
        else:
            osr_set_traditional_axis_mapping_strategy(obj._osr)
            return obj

    @staticmethod
    def from_dict(initialdata=None, **kwargs):
        """Make a CRS from a dict of PROJ parameters or PROJ JSON.

        Parameters
        ----------
        initialdata : mapping, optional
            A dictionary or other mapping
        kwargs : mapping, optional
            Another mapping. Will be overlaid on the initialdata.

        Returns
        -------
        CRS

        Raises
        ------
        CRSError

        """
        cdef const char *text_c = NULL
        cdef CRS obj

        if initialdata is not None:
            data = dict(initialdata.items())
        else:
            data = {}
        data.update(**kwargs)

        # "+init=epsg:xxxx" is deprecated in GDAL. If we find this, we will
        # extract the epsg code and dispatch to from_epsg.
        if 'init' in data and data['init'].lower().startswith('epsg:'):
            epsg_code = int(data['init'].split(':')[1])
            return CRS.from_epsg(epsg_code)

        elif not ("init" in data or "proj" in data):
            # We've been given a PROJ JSON-encoded text.
            text_b = json.dumps(data).encode('utf-8')
            text_c = text_b
            obj = CRS.__new__(CRS)
            try:
                errcode = exc_wrap_ogrerr(OSRSetFromUserInput(obj._osr, text_c))
            except CPLE_BaseError as exc:
                raise CRSError("The WKT could not be parsed. {}".format(exc))
            else:
                osr_set_traditional_axis_mapping_strategy(obj._osr)
                obj._data = data
                return obj

        else:
            # Continue with the general case.
            pjargs = []
            for key in data.keys() & all_proj_keys:
                val = data[key]
                if val is None or val is True:
                    pjargs.append('+{}'.format(key))
                elif val is False:
                    pass
                else:
                    pjargs.append('+{}={}'.format(key, val))

            proj = ' '.join(pjargs)
            b_proj = proj.encode('utf-8')
            obj = CRS.__new__(CRS)

            try:
                exc_wrap_ogrerr(OSRImportFromProj4(obj._osr, <const char *>b_proj))
            except CPLE_BaseError as exc:
                raise CRSError("The PROJ4 dict could not be understood. {}".format(exc))
            else:
                osr_set_traditional_axis_mapping_strategy(obj._osr)
                obj._data = data
                return obj

    @staticmethod
    def from_wkt(wkt, morph_from_esri_dialect=False):
        """Make a CRS from a WKT string.

        Parameters
        ----------
        wkt : str
            A WKT string.
        morph_from_esri_dialect : bool, optional
            If True, items in the input using Esri's dialect of WKT
            will be replaced by OGC standard equivalents.

        Returns
        -------
        CRS

        Raises
        ------
        CRSError

        """
        cdef char *wkt_c = NULL

        if not isinstance(wkt, str):
            raise ValueError("A string is expected")

        wkt_b= wkt.encode('utf-8')
        wkt_c = wkt_b

        cdef CRS obj = CRS.__new__(CRS)

        try:
            errcode = exc_wrap_ogrerr(OSRImportFromWkt(obj._osr, &wkt_c))
        except CPLE_BaseError as exc:
            raise CRSError("The WKT could not be parsed. {}".format(exc))
        else:
            osr_set_traditional_axis_mapping_strategy(obj._osr)
            return obj

    @staticmethod
    def from_user_input(value, morph_from_esri_dialect=False):
        """Make a CRS from a variety of inputs.

        Parameters
        ----------
        value : object
            User input of many different kinds.
        morph_from_esri_dialect : bool, optional
            If True, items in the input using Esri's dialect of WKT
            will be replaced by OGC standard equivalents.

        Returns
        -------
        CRS

        Raises
        ------
        CRSError

        """
        cdef const char *text_c = NULL
        cdef CRS obj

        if isinstance(value, CRS):
            return value
        elif hasattr(value, "to_wkt") and callable(value.to_wkt):
            return CRS.from_wkt(value.to_wkt(), morph_from_esri_dialect=morph_from_esri_dialect)
        elif isinstance(value, int):
            return CRS.from_epsg(value)
        elif isinstance(value, dict):
            return CRS.from_dict(value)
        elif isinstance(value, str):
            return CRS.from_string(value)
        else:
            raise CRSError("CRS is invalid: {!r}".format(value))

    @staticmethod
    def from_authority(auth_name, code):
        """Make a CRS from an authority name and code.

        .. versionadded:: 1.1.7

        Parameters
        ----------
        auth_name: str
        code : int or str
            The code used by the authority.

        Returns
        -------
        CRS

        Raises
        ------
        CRSError

        """
        return CRS.from_string("{auth_name}:{code}".format(auth_name=auth_name, code=code))

    @staticmethod
    def from_string(value, morph_from_esri_dialect=False):
        """Make a CRS from an EPSG, PROJ, or WKT string

        Parameters
        ----------
        value : str
            An EPSG, PROJ, or WKT string.
        morph_from_esri_dialect : bool, optional
            If True, items in the input using Esri's dialect of WKT
            will be replaced by OGC standard equivalents.

        Returns
        -------
        CRS

        Raises
        ------
        CRSError
        """
        cdef const char *text_c = NULL
        cdef CRS obj

        try:
            value = value.strip()
        except AttributeError:
            pass

        if not value:
            raise CRSError("CRS is empty or invalid: {!r}".format(value))

        elif value.upper().startswith('EPSG:') and "+" not in value:
            auth, val = value.split(':')
            if not val:
                raise CRSError("Invalid CRS: {!r}".format(value))
            return CRS.from_epsg(val)

        elif value.startswith('{') or value.startswith('['):
            # may be json, try to decode it
            try:
                val = json.loads(value, strict=False)
            except ValueError:
                raise CRSError('CRS appears to be JSON but is not valid')

            if not val:
                raise CRSError("CRS is empty JSON")
            else:
                return CRS.from_dict(**val)

        elif value.endswith("]"):
            return CRS.from_wkt(value, morph_from_esri_dialect=morph_from_esri_dialect)
        elif "=" in value:
            return CRS.from_proj4(value)
        else:
            text_b = value.encode('utf-8')
            text_c = text_b
            obj = CRS.__new__(CRS)
            try:
                errcode = exc_wrap_ogrerr(OSRSetFromUserInput(obj._osr, text_c))
            except CPLE_BaseError as exc:
                raise CRSError("The WKT could not be parsed. {}".format(exc))
            else:
                osr_set_traditional_axis_mapping_strategy(obj._osr)
                return obj

    def __cinit__(self):
        self._osr = OSRNewSpatialReference(NULL)

    def __dealloc__(self):
        _safe_osr_release(self._osr)

    def __hash__(self):
        return hash(self.wkt)

    def __getitem__(self, item):
        return self.data[item]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def get(self, item):
        return self.data.get(item)

    def items(self):
        return self.data.items()

    def __bool__(self):
        return bool(self.wkt)

    __nonzero__ = __bool__

    def __getstate__(self):
        return self.to_wkt()

    def __setstate__(self, state):
        cdef CRS tmp
        tmp = CRS.from_wkt(state)
        self._osr = OSRClone(tmp._osr)
        self._wkt = tmp._wkt
        self._data = tmp.data
        self._epsg = tmp._epsg

    def __copy__(self):
        return pickle.loads(pickle.dumps(self))

    def __hash__(self):
        return hash(self.to_wkt())

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        if self._epsg:
            return "CRS.from_epsg({})".format(self._epsg)
        else:
            return "CRS.from_wkt('{}')".format(self.wkt)

    def __eq__(self, other):
        return self.equals(other, ignore_axis_order=False)

    def _projjson(self):
        """Get a PROJ JSON representation.

        For internal use only.

        .. versionadded:: 1.3.0

        .. note:: Requires GDAL 3.1+ and PROJ 6.2+

        Returns
        -------
        projjson : str
            PROJ JSON-encoded text.

        Raises
        ------
        CRSError

        """
        cdef char *conv_json = NULL
        cdef const char* options[2]

        try:
            if OSRGetName(self._osr) != NULL:
                options[0] = b"MULTILINE=NO"
                options[1] = NULL
                exc_wrap_ogrerr(OSRExportToPROJJSON(self._osr, &conv_json, options))
        except CPLE_BaseError as exc:
            raise CRSError("Cannot convert to PROJ JSON. {}".format(exc))

        else:
            if conv_json != NULL:
                return conv_json.decode('utf-8')
            else:
                return ''
        finally:
            CPLFree(conv_json)


def epsg_treats_as_latlong(input_crs):
    """Test if the CRS is in latlon order

    From GDAL docs:

    > This method returns TRUE if EPSG feels this geographic coordinate
    system should be treated as having lat/long coordinate ordering.

    > Currently this returns TRUE for all geographic coordinate systems with
    an EPSG code set, and axes set defining it as lat, long.

    > FALSE will be returned for all coordinate systems that are not
    geographic, or that do not have an EPSG code set.

    > **Note**

    > Important change of behavior since GDAL 3.0.
    In previous versions, geographic CRS imported with importFromEPSG()
    would cause this method to return FALSE on them, whereas now it returns
    TRUE, since importFromEPSG() is now equivalent to importFromEPSGA().

    Parameters
    ----------
    input_crs : CRS
        Coordinate reference system, as a rasterio CRS object
        Example: CRS({'init': 'EPSG:4326'})

    Returns
    -------
    bool

    """
    cdef CRS crs

    if not isinstance(input_crs, CRS):
        crs = CRS.from_user_input(input_crs)
    else:
        crs = input_crs

    try:
        return bool(OSREPSGTreatsAsLatLong(crs._osr) == 1)
    except CPLE_BaseError as exc:
        raise CRSError("{}".format(exc))


def epsg_treats_as_northingeasting(input_crs):
    """Test if the CRS should be treated as having northing/easting coordinate ordering

    From GDAL docs:

    > This method returns TRUE if EPSG feels this projected coordinate
    system should be treated as having northing/easting coordinate ordering.

    > Currently this returns TRUE for all projected coordinate systems with
    an EPSG code set, and axes set defining it as northing, easting.

    > FALSE will be returned for all coordinate systems that are not
    projected, or that do not have an EPSG code set.

    > **Note**

    > Important change of behavior since GDAL 3.0.
    In previous versions, projected CRS with northing, easting axis order
    imported with importFromEPSG() would cause this method to return FALSE
    on them, whereas now it returns TRUE, since importFromEPSG() is now 
    equivalent to importFromEPSGA().

    Parameters
    ----------
    input_crs : CRS
        Coordinate reference system, as a rasterio CRS object
        Example: CRS({'init': 'EPSG:4326'})

    Returns
    -------
    bool

    """
    cdef CRS crs

    if not isinstance(input_crs, CRS):
        crs = CRS.from_user_input(input_crs)
    else:
        crs = input_crs

    try:
        return bool(OSREPSGTreatsAsNorthingEasting(crs._osr) == 1)
    except CPLE_BaseError as exc:
        raise CRSError("{}".format(exc))


# Below is the big list of PROJ4 parameters from
# http://trac.osgeo.org/proj/wiki/GenParms.
# It is parsed into a list of parameter keys ``all_proj_keys``.

_param_data = """
+a         Semimajor radius of the ellipsoid axis
+alpha     ? Used with Oblique Mercator and possibly a few others
+axis      Axis orientation (new in 4.8.0)
+b         Semiminor radius of the ellipsoid axis
+datum     Datum name (see `proj -ld`)
+ellps     Ellipsoid name (see `proj -le`)
+init      Initialize from a named CRS
+k         Scaling factor (old name)
+k_0       Scaling factor (new name)
+lat_0     Latitude of origin
+lat_1     Latitude of first standard parallel
+lat_2     Latitude of second standard parallel
+lat_ts    Latitude of true scale
+lon_0     Central meridian
+lonc      ? Longitude used with Oblique Mercator and possibly a few others
+lon_wrap  Center longitude to use for wrapping (see below)
+nadgrids  Filename of NTv2 grid file to use for datum transforms (see below)
+no_defs   Don't use the /usr/share/proj/proj_def.dat defaults file
+over      Allow longitude output outside -180 to 180 range, disables wrapping (see below)
+pm        Alternate prime meridian (typically a city name, see below)
+proj      Projection name (see `proj -l`)
+south     Denotes southern hemisphere UTM zone
+to_meter  Multiplier to convert map units to 1.0m
+towgs84   3 or 7 term datum transform parameters (see below)
+units     meters, US survey feet, etc.
+vto_meter vertical conversion to meters.
+vunits    vertical units.
+x_0       False easting
+y_0       False northing
+zone      UTM zone
+a         Semimajor radius of the ellipsoid axis
+alpha     ? Used with Oblique Mercator and possibly a few others
+azi
+b         Semiminor radius of the ellipsoid axis
+belgium
+beta
+czech
+e         Eccentricity of the ellipsoid = sqrt(1 - b^2/a^2) = sqrt( f*(2-f) )
+ellps     Ellipsoid name (see `proj -le`)
+es        Eccentricity of the ellipsoid squared
+f         Flattening of the ellipsoid (often presented as an inverse, e.g. 1/298)
+gamma
+geoc
+guam
+h
+k         Scaling factor (old name)
+K
+k_0       Scaling factor (new name)
+lat_0     Latitude of origin
+lat_1     Latitude of first standard parallel
+lat_2     Latitude of second standard parallel
+lat_b
+lat_t
+lat_ts    Latitude of true scale
+lon_0     Central meridian
+lon_1
+lon_2
+lonc      ? Longitude used with Oblique Mercator and possibly a few others
+lsat
+m
+M
+n
+no_cut
+no_off
+no_rot
+ns
+o_alpha
+o_lat_1
+o_lat_2
+o_lat_c
+o_lat_p
+o_lon_1
+o_lon_2
+o_lon_c
+o_lon_p
+o_proj
+over
+p
+path
+proj      Projection name (see `proj -l`)
+q
+R
+R_a
+R_A       Compute radius such that the area of the sphere is the same as the area of the ellipsoid
+rf        Reciprocal of the ellipsoid flattening term (e.g. 298)
+R_g
+R_h
+R_lat_a
+R_lat_g
+rot
+R_V
+s
+south     Denotes southern hemisphere UTM zone
+sym
+t
+theta
+tilt
+to_meter  Multiplier to convert map units to 1.0m
+units     meters, US survey feet, etc.
+vopt
+W
+westo
+wktext
+x_0       False easting
+y_0       False northing
+zone      UTM zone
"""

all_proj_keys = set(line.split(' ', 1)[0][1:] for line in filter(None, _param_data.splitlines()))
all_proj_keys.add('no_mayo')



"""Abstraction for sessions in various clouds."""

import logging
import os
from types import SimpleNamespace

from rasterio._path import _parse_path, _UnparsedPath


log = logging.getLogger(__name__)

try:
    import boto3
except ImportError:
    log.debug("Could not import boto3, continuing with reduced functionality.")
    boto3 = None


class Session:
    """Base for classes that configure access to secured resources.

    Attributes
    ----------
    credentials : dict
        Keys and values for session credentials.

    Notes
    -----
    This class is not intended to be instantiated.

    """

    @classmethod
    def hascreds(cls, config):
        """Determine if the given configuration has proper credentials

        Parameters
        ----------
        cls : class
            A Session class.
        config : dict
            GDAL configuration as a dict.

        Returns
        -------
        bool

        """
        return NotImplemented

    def get_credential_options(self):
        """Get credentials as GDAL configuration options

        Returns
        -------
        dict

        """
        return NotImplemented

    @staticmethod
    def from_foreign_session(session, cls=None):
        """Create a session object matching the foreign `session`.

        Parameters
        ----------
        session : obj
            A foreign session object.
        cls : Session class, optional
            The class to return.

        Returns
        -------
        Session

        """
        if not cls:
            return DummySession()
        else:
            return cls(session)

    @staticmethod
    def cls_from_path(path):
        """Find the session class suited to the data at `path`.

        Parameters
        ----------
        path : str
            A dataset path or identifier.

        Returns
        -------
        class

        """
        if not path:
            return DummySession

        path = _parse_path(path)

        if isinstance(path, _UnparsedPath) or path.is_local:
            return DummySession

        elif (
            path.scheme == "s3" or "amazonaws.com" in path.path
        ) and "X-Amz-Signature" not in path.path:
            if boto3 is not None:
                return AWSSession
            else:
                log.info("boto3 not available, falling back to a DummySession.")
                return DummySession

        elif path.scheme == "oss" or "aliyuncs.com" in path.path:
            return OSSSession

        elif path.path.startswith("/vsiswift/"):
            return SwiftSession

        elif path.scheme == "az":
            return AzureSession

        # This factory can be extended to other cloud providers here.
        # elif path.scheme == "cumulonimbus":  # for example.
        #     return CumulonimbusSession(*args, **kwargs)

        else:
            return DummySession

    @staticmethod
    def from_path(path, *args, **kwargs):
        """Create a session object suited to the data at `path`.

        Parameters
        ----------
        path : str
            A dataset path or identifier.
        args : sequence
            Positional arguments for the foreign session constructor.
        kwargs : dict
            Keyword arguments for the foreign session constructor.

        Returns
        -------
        Session

        """
        return Session.cls_from_path(path)(*args, **kwargs)

    @staticmethod
    def aws_or_dummy(*args, **kwargs):
        """Create an AWSSession if boto3 is available, else DummySession

        Parameters
        ----------
        path : str
            A dataset path or identifier.
        args : sequence
            Positional arguments for the foreign session constructor.
        kwargs : dict
            Keyword arguments for the foreign session constructor.

        Returns
        -------
        Session

        """
        if boto3 is not None:
            return AWSSession(*args, **kwargs)
        else:
            return DummySession(*args, **kwargs)

    @staticmethod
    def from_environ(*args, **kwargs):
        """Create a session object suited to the environment.

        Parameters
        ----------
        path : str
            A dataset path or identifier.
        args : sequence
            Positional arguments for the foreign session constructor.
        kwargs : dict
            Keyword arguments for the foreign session constructor.

        Returns
        -------
        Session

        """
        try:
            session = Session.aws_or_dummy(*args, **kwargs)
            session.credentials
        except RuntimeError:
            log.warning("Credentials in environment have expired. Creating a DummySession.")
            session = DummySession(*args, **kwargs)
        return session


class DummySession(Session):
    """A dummy session.

    Attributes
    ----------
    credentials : dict
        The session credentials.

    """

    def __init__(self, *args, **kwargs):
        self._session = None
        self.credentials = {}

    @classmethod
    def hascreds(cls, config):
        """Determine if the given configuration has proper credentials

        Parameters
        ----------
        cls : class
            A Session class.
        config : dict
            GDAL configuration as a dict.

        Returns
        -------
        bool

        """
        return True

    def get_credential_options(self):
        """Get credentials as GDAL configuration options

        Returns
        -------
        dict

        """
        return {}


class AWSSession(Session):
    """Configures access to secured resources stored in AWS S3.
    """

    def __init__(
            self, session=None, aws_unsigned=None, aws_access_key_id=None,
            aws_secret_access_key=None, aws_session_token=None,
            region_name=None, profile_name=None, endpoint_url=None,
            requester_pays=False):
        """Create a new AWS session

        Parameters
        ----------
        session : optional
            A boto3 session object.
        aws_unsigned : bool, optional (default: False)
            If True, requests will be unsigned.
        aws_access_key_id : str, optional
            An access key id, as per boto3.
        aws_secret_access_key : str, optional
            A secret access key, as per boto3.
        aws_session_token : str, optional
            A session token, as per boto3.
        region_name : str, optional
            A region name, as per boto3.
        profile_name : str, optional
            A shared credentials profile name, as per boto3.
        endpoint_url: str, optional
            An endpoint_url, as per GDAL's AWS_S3_ENPOINT
        requester_pays : bool, optional
            True if the requester agrees to pay transfer costs (default:
            False)
        """
        if aws_unsigned is None:
            aws_unsigned = parse_bool(os.getenv("AWS_NO_SIGN_REQUEST", False))

        if session:
            self._session = session
        elif aws_unsigned:
            self._session = SimpleNamespace(region_name=region_name)
        else:
            self._session = boto3.Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=region_name,
                profile_name=profile_name)

        self.requester_pays = requester_pays
        self.unsigned = aws_unsigned
        self.endpoint_url = endpoint_url
        self._creds = (
            self._session.get_credentials()
            if not self.unsigned
            else None
        )

    @classmethod
    def hascreds(cls, config):
        """Determine if the given configuration has proper credentials

        Parameters
        ----------
        cls : class
            A Session class.
        config : dict
            GDAL configuration as a dict.

        Returns
        -------
        bool

        """
        return ('AWS_ACCESS_KEY_ID' in config and 'AWS_SECRET_ACCESS_KEY' in config) or 'AWS_NO_SIGN_REQUEST' in config

    @property
    def credentials(self):
        """The session credentials as a dict"""
        res = {}
        if self._creds:  # pragma: no branch
            frozen_creds = self._creds.get_frozen_credentials()
            if frozen_creds.access_key:  # pragma: no branch
                res['aws_access_key_id'] = frozen_creds.access_key
            if frozen_creds.secret_key:  # pragma: no branch
                res['aws_secret_access_key'] = frozen_creds.secret_key
            if frozen_creds.token:
                res['aws_session_token'] = frozen_creds.token
        if self._session.region_name:
            res['aws_region'] = self._session.region_name
        if self.requester_pays:
            res['aws_request_payer'] = 'requester'
        if self.endpoint_url:
            res['aws_s3_endpoint'] = self.endpoint_url
        return res

    def get_credential_options(self):
        """Get credentials as GDAL configuration options

        Returns
        -------
        dict

        """
        if self.unsigned:
            opts = {'AWS_NO_SIGN_REQUEST': 'YES'}
            opts.update({k.upper(): v for k, v in self.credentials.items()
                        if k in ('aws_region', 'aws_s3_endpoint')})
            return opts
        else:
            return {k.upper(): v for k, v in self.credentials.items()}


class OSSSession(Session):
    """Configures access to secured resources stored in Alibaba Cloud OSS.
    """
    def __init__(self, oss_access_key_id=None, oss_secret_access_key=None, oss_endpoint=None):
        """Create new Alibaba Cloud OSS session

        Parameters
        ----------
        oss_access_key_id: string, optional (default: None)
            An access key id
        oss_secret_access_key: string, optional (default: None)
            An secret access key
        oss_endpoint: string, optional (default: None)
            the region attached to the bucket
        """

        self._creds = {
            "oss_access_key_id": oss_access_key_id,
            "oss_secret_access_key": oss_secret_access_key,
            "oss_endpoint": oss_endpoint
        }

    @classmethod
    def hascreds(cls, config):
        """Determine if the given configuration has proper credentials

        Parameters
        ----------
        cls : class
            A Session class.
        config : dict
            GDAL configuration as a dict.

        Returns
        -------
        bool

        """
        return 'OSS_ACCESS_KEY_ID' in config and 'OSS_SECRET_ACCESS_KEY' in config

    @property
    def credentials(self):
        """The session credentials as a dict"""
        return self._creds

    def get_credential_options(self):
        """Get credentials as GDAL configuration options

        Returns
        -------
        dict

        """
        return {k.upper(): v for k, v in self.credentials.items()}


class GSSession(Session):
    """Configures access to secured resources stored in Google Cloud Storage
    """
    def __init__(self, google_application_credentials=None):
        """Create new Google Cloud Storage session

        Parameters
        ----------
        google_application_credentials: string
            Path to the google application credentials JSON file.
        """

        if google_application_credentials is not None:
            self._creds = {'google_application_credentials': google_application_credentials}
        else:
            self._creds = {}

    @classmethod
    def hascreds(cls, config):
        """Determine if the given configuration has proper credentials

        Parameters
        ----------
        cls : class
            A Session class.
        config : dict
            GDAL configuration as a dict.

        Returns
        -------
        bool

        """
        return 'GOOGLE_APPLICATION_CREDENTIALS' in config

    @property
    def credentials(self):
        """The session credentials as a dict"""
        return self._creds

    def get_credential_options(self):
        """Get credentials as GDAL configuration options

        Returns
        -------
        dict

        """
        return {k.upper(): v for k, v in self.credentials.items()}


class SwiftSession(Session):
    """Configures access to secured resources stored in OpenStack Swift Object Storage.
    """
    def __init__(self, session=None,
                swift_storage_url=None, swift_auth_token=None,
                swift_auth_v1_url=None, swift_user=None, swift_key=None):
        """Create new OpenStack Swift Object Storage Session.

        Three methods are possible:
            1. Create session by the swiftclient library.
            2. The SWIFT_STORAGE_URL and SWIFT_AUTH_TOKEN (this method is recommended by GDAL docs).
            3. The SWIFT_AUTH_V1_URL, SWIFT_USER and SWIFT_KEY (This depends on the swiftclient library).

        Parameters
        ----------
        session: optional
            A swiftclient connection object
        swift_storage_url:
            the storage URL
        swift_auth_token:
            the value of the x-auth-token authorization token
        swift_storage_url: string, optional
            authentication URL
        swift_user: string, optional
            user name to authenticate as
        swift_key: string, optional
            key/password to authenticate with

        Examples
        --------
        >>> import rasterio
        >>> from rasterio.session import SwiftSession
        >>> fp = '/vsiswift/bucket/key.tif'
        >>> conn = Connection(authurl='http://127.0.0.1:7777/auth/v1.0', user='test:tester', key='testing')
        >>> session = SwiftSession(conn)
        >>> with rasterio.Env(session):
        >>>     with rasterio.open(fp) as src:
        >>>         print(src.profile)

        """
        if swift_storage_url and swift_auth_token:
            self._creds = {
                "swift_storage_url": swift_storage_url,
                "swift_auth_token": swift_auth_token
            }
        else:
            from swiftclient.client import Connection

            if session:
                self._session = session
            else:
                self._session = Connection(
                    authurl=swift_auth_v1_url,
                    user=swift_user,
                    key=swift_key
                )
            self._creds = {
                "swift_storage_url": self._session.get_auth()[0],
                "swift_auth_token": self._session.get_auth()[1]
            }

    @classmethod
    def hascreds(cls, config):
        """Determine if the given configuration has proper credentials
        Parameters
        ----------
        cls : class
            A Session class.
        config : dict
            GDAL configuration as a dict.
        Returns
        -------
        bool
        """
        return 'SWIFT_STORAGE_URL' in config and 'SWIFT_AUTH_TOKEN' in config

    @property
    def credentials(self):
        """The session credentials as a dict"""
        return self._creds

    def get_credential_options(self):
        """Get credentials as GDAL configuration options
        Returns
        -------
        dict
        """
        return {k.upper(): v for k, v in self.credentials.items()}


class AzureSession(Session):
    """Configures access to secured resources stored in Microsoft Azure Blob Storage.
    """
    def __init__(self, azure_storage_connection_string=None,
                 azure_storage_account=None, azure_storage_access_key=None,
                 azure_unsigned=False):
        """Create new Microsoft Azure Blob Storage session

        Parameters
        ----------
        azure_storage_connection_string: string
            A connection string contains both an account name and a secret key.
        azure_storage_account: string
            An account name
        azure_storage_access_key: string
            A secret key
        azure_unsigned : bool, optional (default: False)
            If True, requests will be unsigned.
        """

        self.unsigned = parse_bool(os.getenv("AZURE_NO_SIGN_REQUEST", azure_unsigned))
        self.storage_account = azure_storage_account or os.getenv("AZURE_STORAGE_ACCOUNT")
        self.storage_access_key = azure_storage_access_key or os.getenv("AZURE_STORAGE_ACCESS_KEY")

        if azure_storage_connection_string:
            self._creds = {
                "azure_storage_connection_string": azure_storage_connection_string
            }
        elif not self.unsigned:
            self._creds = {
                "azure_storage_account": self.storage_account,
                "azure_storage_access_key": self.storage_access_key
            }
        else:
            self._creds = {
                "azure_storage_account": self.storage_account
            }

    @classmethod
    def hascreds(cls, config):
        """Determine if the given configuration has proper credentials

        Parameters
        ----------
        cls : class
            A Session class.
        config : dict
            GDAL configuration as a dict.

        Returns
        -------
        bool

        """
        return (
            'AZURE_STORAGE_CONNECTION_STRING' in config
            or ('AZURE_STORAGE_ACCOUNT' in config and 'AZURE_STORAGE_ACCESS_KEY' in config)
            or ('AZURE_STORAGE_ACCOUNT' in config and 'AZURE_NO_SIGN_REQUEST' in config)
        )

    @property
    def credentials(self):
        """The session credentials as a dict"""
        return self._creds

    def get_credential_options(self):
        """Get credentials as GDAL configuration options

        Returns
        -------
        dict

        """
        if self.unsigned:
            return {
                'AZURE_NO_SIGN_REQUEST': 'YES',
                'AZURE_STORAGE_ACCOUNT': self.storage_account
            }
        else:
            return {k.upper(): v for k, v in self.credentials.items()}

def parse_bool(v):
    """CPLTestBool equivalent"""
    if isinstance(v, bool):
        return v
    if isinstance(v, str):
        return v.lower() not in ("no", "false", "off", "0")
    return bool(v)



from hypothesis import given
from hypothesis.strategies import composite, floats, integers

from rasterio import windows
from rasterio.coords import BoundingBox


class FakeDataset(windows.WindowMethodsMixin):
    """A test double to support testing of the mixin class"""

    def __init__(self, **kwargs):
        self.transform = kwargs.get('transform')
        self.height = kwargs.get('height')
        self.width = kwargs.get('width')

    @property
    def bounds(self):
        a, b, c, d, e, f, _, _, _ = self.transform
        return BoundingBox(c, f + e * self.height, c + a * self.width, f)

    def __repr__(self):
        return f"<FakeDataset: transform={self.transform}, height={self.height}, width={self.width}>"


@composite
def dataset_utm(draw):
    """Generate a fake UTM dataset with an origin, a resolution, and a finite size"""
    x = draw(floats(min_value=-1e6, max_value=1e+6, allow_nan=False, allow_infinity=False))
    y = draw(floats(min_value=-1e6, max_value=1e+6, allow_nan=False, allow_infinity=False))
    res = draw(floats(min_value=0.1, max_value=30, allow_nan=False, allow_infinity=False))
    h = draw(integers(min_value=1, max_value=1000))
    w = draw(integers(min_value=1, max_value=1000))
    return FakeDataset(
        transform=windows.Affine.identity() * windows.Affine.translation(x, y) * windows.Affine.scale(res, -res),
        height=h, width=w)


@composite
def dataset_utm_north_down(draw):
    """Generate a fake UTM dataset with an origin, a resolution, and a finite size"""
    x = draw(floats(min_value=-1e6, max_value=1e+6, allow_nan=False, allow_infinity=False))
    y = draw(floats(min_value=-1e6, max_value=1e+6, allow_nan=False, allow_infinity=False))
    res = draw(floats(min_value=0.1, max_value=30, allow_nan=False, allow_infinity=False))
    h = draw(integers(min_value=1, max_value=1000))
    w = draw(integers(min_value=1, max_value=1000))
    return FakeDataset(
        transform=windows.Affine.identity() * windows.Affine.translation(x, y) * windows.Affine.scale(res),
        height=h, width=w)


def assert_windows_almost_equal(a, b):
    assert round(a.col_off, 7) == round(b.col_off, 7)
    assert round(a.row_off, 7) == round(b.row_off, 7)
    assert round(a.width, 7) == round(b.width, 7)
    assert round(a.height, 7) == round(b.height, 7)


@given(dataset=dataset_utm())
def test_window_rt(dataset):
    """Get correct window for full dataset extent"""
    left, top = dataset.transform * (0, 0)
    right, bottom = dataset.transform * (dataset.width, dataset.height)
    assert_windows_almost_equal(
        dataset.window(left, bottom, right, top),
        windows.Window(0, 0, dataset.width, dataset.height))


@given(dataset=dataset_utm_north_down())
def test_window_rt_north_down(dataset):
    """Get correct window for full dataset extent"""
    left, top = dataset.transform * (0, 0)
    right, bottom = dataset.transform * (dataset.width, dataset.height)
    assert_windows_almost_equal(
        dataset.window(left, bottom, right, top),
        windows.Window(0, 0, dataset.width, dataset.height))


@given(dataset=dataset_utm())
def test_window_transform_rt(dataset):
    assert dataset.window_transform(windows.Window(0, 0, dataset.width, dataset.height)) == dataset.transform


@given(dataset=dataset_utm_north_down())
def test_window_transform_rt_north_down(dataset):
    assert dataset.window_transform(windows.Window(0, 0, dataset.width, dataset.height)) == dataset.transform


@given(dataset=dataset_utm())
def test_window_bounds_rt(dataset):
    assert dataset.window_bounds(windows.Window(0, 0, dataset.width, dataset.height)) == dataset.bounds


@given(dataset=dataset_utm_north_down())
def test_window_bounds_rt_north_down(dataset):
    assert dataset.window_bounds(windows.Window(0, 0, dataset.width, dataset.height)) == dataset.bounds



"""$ rio calc"""

from collections import OrderedDict, UserDict
from contextlib import ExitStack
import math
from collections.abc import Mapping

import click
import numpy

import rasterio
from rasterio._vendor import snuggs
from rasterio.features import sieve
from rasterio.fill import fillnodata
from rasterio.rio import options
from rasterio.rio.helpers import resolve_inout
from rasterio.windows import Window, subdivide


def _get_bands(inputs, sources, d, i=None):
    """Get a rasterio.Band object from calc's inputs."""
    idx = d if d in dict(inputs) else int(d) - 1
    src = sources[idx]
    return rasterio.band(src, i) if i else [rasterio.band(src, j) for j in src.indexes]


def _read_array(ix, subix=None, dtype=None):
    """Change the type of a read array."""
    arr = snuggs._ctx.lookup(ix, subix)
    if dtype:
        arr = arr.astype(dtype)
    return arr


def asarray(*args):
    if len(args) == 1 and hasattr(args[0], "__iter__"):
        return numpy.asanyarray(list(args[0]))
    else:
        return numpy.asanyarray(list(args))


class FuncMapper(UserDict, Mapping):
    """Resolves functions from names in pipeline expressions."""

    def __getitem__(self, key):
        """Get a function by its name."""
        if key in self.data:
            return self.data[key]
        elif key in __builtins__ and not key.startswith("__"):
            return __builtins__[key]
        else:
            return (
                lambda g, *args, **kwargs: getattr(g, key)(*args, **kwargs)
                if callable(getattr(g, key))
                else getattr(g, key)
            )


@click.command(short_help="Raster data calculator.")
@click.argument('command')
@options.files_inout_arg
@options.output_opt
@options.format_opt
@click.option('--name', multiple=True,
              help='Specify an input file with a unique short (alphas only) '
                   'name for use in commands like '
                   '"a=tests/data/RGB.byte.tif".')
@options.dtype_opt
@options.masked_opt
@options.overwrite_opt
@click.option("--mem-limit", type=int, default=64, help="Limit on memory used to perform calculations, in MB.")
@options.creation_options
@click.pass_context
def calc(ctx, command, files, output, driver, name, dtype, masked, overwrite, mem_limit, creation_options):
    """A raster data calculator

    Evaluates an expression using input datasets and writes the result
    to a new dataset.

    Command syntax is lisp-like. An expression consists of an operator
    or function name and one or more strings, numbers, or expressions
    enclosed in parentheses. Functions include ``read`` (gets a raster
    array) and ``asarray`` (makes a 3-D array from 2-D arrays).

    \b
        * (read i) evaluates to the i-th input dataset (a 3-D array).
        * (read i j) evaluates to the j-th band of the i-th dataset (a
          2-D array).
        * (read i j 'float64') casts the array to, e.g. float64. This
          is critical if calculations will produces values that exceed
          the limits of the dataset's natural data type.
        * (take foo j) evaluates to the j-th band of a dataset named foo
          (see help on the --name option above).
        * Standard numpy array operators (+, -, \*, /) are available.
        * When the final result is a list of arrays, a multiple band
          output file is written.
        * When the final result is a single array, a single band output
          file is written.

    Example:

    \b
         $ rio calc "(+ 2 (* 0.95 (read 1)))" tests/data/RGB.byte.tif \\
         > /tmp/out.tif

    The command above produces a 3-band GeoTIFF with all values scaled
    by 0.95 and incremented by 2.

    \b
        $ rio calc "(asarray (+ 125 (read 1)) (read 1) (read 1))" \\
        > tests/data/shade.tif /tmp/out.tif

    The command above produces a 3-band RGB GeoTIFF, with red levels
    incremented by 125, from the single-band input.

    The maximum amount of memory used to perform calculations defaults to
    64 MB. This number can be increased to improve speed of calculation.

    """  # noqa: W605
    dst = None
    sources = []

    with ctx.obj["env"], ExitStack() as stack:
        output, files = resolve_inout(files=files, output=output, overwrite=overwrite)
        inputs = [tuple(n.split("=")) for n in name] + [(None, n) for n in files]
        sources = [stack.enter_context(rasterio.open(path)) for name, path in inputs]

        snuggs.func_map = FuncMapper(
            asarray=asarray,
            take=lambda a, idx: numpy.take(a, idx - 1, axis=0),
            read=_read_array,
            band=lambda d, i: _get_bands(inputs, sources, d, i),
            bands=lambda d: _get_bands(inputs, sources, d),
            fillnodata=fillnodata,
            sieve=sieve,
        )

        first = sources[0]
        kwargs = first.profile
        kwargs.update(**creation_options)
        dtype = dtype or first.meta["dtype"]
        kwargs["dtype"] = dtype
        kwargs.pop("driver", None)
        if driver:
            kwargs["driver"] = driver

        # The windows iterator is initialized with a single sample.
        # The actual work windows will be added in the second
        # iteration of the loop.
        work_windows = [Window(0, 0, 16, 16)]

        for window in work_windows:
            ctxkwds = OrderedDict()

            for i, ((name, path), src) in enumerate(zip(inputs, sources)):
                ctxkwds[name or "_i%d" % (i + 1)] = src.read(
                    masked=masked, window=window
                )

            try:
                res = snuggs.eval(command, **ctxkwds)
            except snuggs.ExpressionError as err:
                click.echo("Expression Error:")
                click.echo(f"  {err.text}")
                click.echo(" {}^".format(" " * err.offset))
                click.echo(err)
                raise click.Abort()
            else:
                results = res.astype(dtype)

            if isinstance(results, numpy.ma.core.MaskedArray):
                results = results.filled(float(kwargs["nodata"]))
                if len(results.shape) == 2:
                    results = numpy.ma.asanyarray([results])
            elif len(results.shape) == 2:
                results = numpy.asanyarray([results])

            # The first iteration is only to get sample results and from them
            # compute some properties of the output dataset.
            if dst is None:
                kwargs["count"] = results.shape[0]
                dst = stack.enter_context(rasterio.open(output, "w", **kwargs))
                max_pixels = mem_limit * 1.0e+6 / (numpy.dtype(dst.dtypes[0]).itemsize * dst.count)
                chunk_size = int(math.floor(math.sqrt(max_pixels)))
                work_windows.extend(
                    subdivide(
                        Window(0, 0, dst.width, dst.height),
                        chunk_size,
                        chunk_size
                    )
                )

            # In subsequent iterations we write results.
            else:
                dst.write(results, window=window)



from rasterio.rio.main import main_group


def test_transform(runner):
    """Coordinates are transformed."""
    result = runner.invoke(
        main_group,
        ["transform", "--dst-crs=EPSG:32618", "--precision=2"],
        input="[-78.0, 23.0]",
    )
    assert result.output.rstrip("\n") == "[192457.13, 2546667.68]"



"""$ rio shapes"""



import logging

import click
import cligj
import rasterio

from rasterio.rio import options
from rasterio.features import dataset_features
from rasterio.rio.helpers import write_features

logger = logging.getLogger(__name__)


@click.command(short_help="Write shapes extracted from bands or masks.")
@options.file_in_arg
@options.output_opt
@cligj.precision_opt
@cligj.indent_opt
@cligj.compact_opt
@cligj.projection_geographic_opt
@cligj.projection_projected_opt
@options.sequence_opt
@cligj.use_rs_opt
@options.geojson_type_opt(allowed=('feature', 'bbox'), default='feature')
@click.option('--band/--mask', default=True,
              help="Choose to extract from a band (the default) or a mask.")
@click.option('--bidx', 'bandidx', type=int, default=None,
              help="Index of the band or mask that is the source of shapes.")
@click.option('--sampling', type=int, default=1,
              help="Inverse of the sampling fraction; "
                   "a value of 10 decimates.")
@click.option('--with-nodata/--without-nodata', default=False,
              help="Include or do not include (the default) nodata regions.")
@click.option('--as-mask/--not-as-mask', default=False,
              help="Interpret a band as a mask and output only one class of "
                   "valid data shapes.")
@click.pass_context
def shapes(
        ctx, input, output, precision, indent, compact, projection, sequence,
        use_rs, geojson_type, band, bandidx, sampling, with_nodata, as_mask):
    """Extracts shapes from one band or mask of a dataset and writes
    them out as GeoJSON. Unless otherwise specified, the shapes will be
    transformed to WGS 84 coordinates.

    The default action of this command is to extract shapes from the
    first band of the input dataset. The shapes are polygons bounding
    contiguous regions (or features) of the same raster value. This
    command performs poorly for int16 or float type datasets.

    Bands other than the first can be specified using the `--bidx`
    option:

      $ rio shapes --bidx 3 tests/data/RGB.byte.tif

    The valid data footprint of a dataset's i-th band can be extracted
    by using the `--mask` and `--bidx` options:

      $ rio shapes --mask --bidx 1 tests/data/RGB.byte.tif

    Omitting the `--bidx` option results in a footprint extracted from
    the conjunction of all band masks. This is generally smaller than
    any individual band's footprint.

    A dataset band may be analyzed as though it were a binary mask with
    the `--as-mask` option:

      $ rio shapes --as-mask --bidx 1 tests/data/RGB.byte.tif
    """
    # These import numpy, which we don't want to do unless it's needed.
    dump_kwds = {'sort_keys': True}
    if indent:
        dump_kwds['indent'] = indent
    if compact:
        dump_kwds['separators'] = (',', ':')

    stdout = click.open_file(
        output, 'w') if output else click.get_text_stream('stdout')

    bidx = 1 if bandidx is None and band else bandidx

    if not sequence:
        geojson_type = 'collection'

    geographic = True if projection == 'geographic' else False

    with ctx.obj["env"] as env:
        with rasterio.open(input) as src:
            write_features(
                stdout,
                feature_gen(
                    src,
                    env,
                    bidx,
                    sampling=sampling,
                    band=band,
                    as_mask=as_mask,
                    with_nodata=with_nodata,
                    geographic=geographic,
                    precision=precision,
                ),
                sequence=sequence,
                geojson_type=geojson_type,
                use_rs=use_rs,
                **dump_kwds
            )


def feature_gen(src, env, *args, **kwargs):
    class Collection:

        def __init__(self, env):
            self.bboxes = []
            self.env = env

        @property
        def bbox(self):
            minxs, minys, maxxs, maxys = zip(*self.bboxes)
            return min(minxs), min(minys), max(maxxs), max(maxys)

        def __call__(self):
            for f in dataset_features(src, *args, **kwargs):
                self.bboxes.append(f['bbox'])
                yield f

    return Collection(env)



# cython: c_string_type=unicode, c_string_encoding=utf8
"""rasterio._err

Exception-raising wrappers for GDAL API functions.

"""

import contextlib
from contextvars import ContextVar
from enum import IntEnum
from itertools import zip_longest
import logging

log = logging.getLogger(__name__)

_ERROR_STACK = ContextVar("error_stack")
_ERROR_STACK.set([])

_GDAL_DEBUG_DOCS = (
    "https://rasterio.readthedocs.io/en/latest/topics/errors.html"
    "#debugging-internal-gdal-functions"
)


class CPLE_BaseError(Exception):
    """Base CPL error class

    Exceptions deriving from this class are intended for use only in
    Rasterio's Cython code. Let's not expose API users to them.
    """

    def __init__(self, error, errno, errmsg):
        self.error = error
        self.errno = errno
        self.errmsg = errmsg

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return "{}".format(self.errmsg)

    @property
    def args(self):
        return self.error, self.errno, self.errmsg


class CPLE_AppDefinedError(CPLE_BaseError):
    pass


class CPLE_OutOfMemoryError(CPLE_BaseError):
    pass


class CPLE_FileIOError(CPLE_BaseError):
    pass


class CPLE_OpenFailedError(CPLE_BaseError):
    pass


class CPLE_IllegalArgError(CPLE_BaseError):
    pass


class CPLE_NotSupportedError(CPLE_BaseError):
    pass


class CPLE_AssertionFailedError(CPLE_BaseError):
    pass


class CPLE_NoWriteAccessError(CPLE_BaseError):
    pass


class CPLE_UserInterruptError(CPLE_BaseError):
    pass


class ObjectNullError(CPLE_BaseError):
    pass


class CPLE_HttpResponseError(CPLE_BaseError):
    pass


class CPLE_AWSBucketNotFoundError(CPLE_BaseError):
    pass


class CPLE_AWSObjectNotFoundError(CPLE_BaseError):
    pass


class CPLE_AWSAccessDeniedError(CPLE_BaseError):
    pass


class CPLE_AWSInvalidCredentialsError(CPLE_BaseError):
    pass


class CPLE_AWSSignatureDoesNotMatchError(CPLE_BaseError):
    pass


class CPLE_AWSError(CPLE_BaseError):
    pass


cdef dict _LEVEL_MAP = {
    0: 0,
    1: logging.DEBUG,
    2: logging.WARNING,
    3: logging.ERROR,
    4: logging.CRITICAL
}

# Map of GDAL error numbers to the Python exceptions.
exception_map = {
    1: CPLE_AppDefinedError,
    2: CPLE_OutOfMemoryError,
    3: CPLE_FileIOError,
    4: CPLE_OpenFailedError,
    5: CPLE_IllegalArgError,
    6: CPLE_NotSupportedError,
    7: CPLE_AssertionFailedError,
    8: CPLE_NoWriteAccessError,
    9: CPLE_UserInterruptError,
    10: ObjectNullError,

    # error numbers 11-16 are introduced in GDAL 2.1. See
    # https://github.com/OSGeo/gdal/pull/98.
    11: CPLE_HttpResponseError,
    12: CPLE_AWSBucketNotFoundError,
    13: CPLE_AWSObjectNotFoundError,
    14: CPLE_AWSAccessDeniedError,
    15: CPLE_AWSInvalidCredentialsError,
    16: CPLE_AWSSignatureDoesNotMatchError,
    17: CPLE_AWSError
}

cdef dict _CODE_MAP = {
    0: 'CPLE_None',
    1: 'CPLE_AppDefined',
    2: 'CPLE_OutOfMemory',
    3: 'CPLE_FileIO',
    4: 'CPLE_OpenFailed',
    5: 'CPLE_IllegalArg',
    6: 'CPLE_NotSupported',
    7: 'CPLE_AssertionFailed',
    8: 'CPLE_NoWriteAccess',
    9: 'CPLE_UserInterrupt',
    10: 'ObjectNull',
    11: 'CPLE_HttpResponse',
    12: 'CPLE_AWSBucketNotFound',
    13: 'CPLE_AWSObjectNotFound',
    14: 'CPLE_AWSAccessDenied',
    15: 'CPLE_AWSInvalidCredentials',
    16: 'CPLE_AWSSignatureDoesNotMatch',
    17: 'CPLE_AWSError'
}

# CPL Error types as an enum.
class GDALError(IntEnum):
    none = CE_None
    debug = CE_Debug
    warning = CE_Warning
    failure = CE_Failure
    fatal = CE_Fatal


cdef inline object exc_check():
    """Checks GDAL error stack for fatal or non-fatal errors

    Returns
    -------
    An Exception, SystemExit, or None

    """
    cdef const char *msg = NULL

    err_type = CPLGetLastErrorType()
    err_no = CPLGetLastErrorNo()
    msg = CPLGetLastErrorMsg()

    if msg == NULL:
        message = "No error message."
    else:
        message = msg
        message = message.replace("`", "'")
        message = message.replace("\n", " ")

    if err_type == 3:
        exception = exception_map.get(err_no, CPLE_BaseError)(err_type, err_no, message)
        CPLErrorReset()
        return exception
    elif err_type == 4:
        exception = SystemExit("Fatal error: {0}".format((err_type, err_no, message)))
        CPLErrorReset()
        return exception
    else:
        CPLErrorReset()
        return None


cdef int exc_wrap(int retval) except -1:
    """Wrap a GDAL function that returns int without checking the retval"""
    exc = exc_check()
    if exc:
        raise exc
    return retval


cdef void log_error(
    CPLErr err_class,
    int err_no,
    const char* msg,
) noexcept with gil:
    """Send CPL errors to Python's logger.

    Because this function is called by GDAL with no Python context, we
    can't propagate exceptions that we might raise here. They'll be
    ignored.

    """
    if err_no in _CODE_MAP:
        # We've observed that some GDAL functions may emit multiple
        # ERROR level messages and yet succeed. We want to see those
        # messages in our log file, but not at the ERROR level. We
        # turn the level down to INFO.
        if err_class == 3:
            log.info(
                "GDAL signalled an error: err_no=%r, msg=%r",
                err_no,
                msg
            )
        elif err_no == 0:
            log.log(_LEVEL_MAP[err_class], "%s", msg)
        else:
            log.log(_LEVEL_MAP[err_class], "%s:%s", _CODE_MAP[err_no], msg)
    else:
        log.info("Unknown error number %r", err_no)


IF UNAME_SYSNAME == "Windows":
    cdef void __stdcall chaining_error_handler(
        CPLErr err_class,
        int err_no,
        const char* msg
    ) noexcept with gil:
        global _ERROR_STACK
        log_error(err_class, err_no, msg)
        if err_class == 3:
            stack = _ERROR_STACK.get()
            stack.append(
                exception_map.get(err_no, CPLE_BaseError)(err_class, err_no, msg),
            )
            _ERROR_STACK.set(stack)
ELSE:
    cdef void chaining_error_handler(
        CPLErr err_class,
        int err_no,
        const char* msg
    ) noexcept with gil:
        global _ERROR_STACK
        log_error(err_class, err_no, msg)
        if err_class == 3:
            stack = _ERROR_STACK.get()
            stack.append(
                exception_map.get(err_no, CPLE_BaseError)(err_class, err_no, msg),
            )
            _ERROR_STACK.set(stack)


cdef int exc_wrap_int(int err) except -1:
    """Wrap a GDAL/OGR function that returns CPLErr or OGRErr (int)

    Raises a Rasterio exception if a non-fatal error has be set.
    """
    if err:
        exc = exc_check()
        if exc:
            raise exc
    CPLErrorReset()
    return err


cdef OGRErr exc_wrap_ogrerr(OGRErr err) except -1:
    """Wrap a function that returns OGRErr but does not use the
    CPL error stack.

    """
    if err == 0:
        CPLErrorReset()
        return err
    else:
        raise CPLE_BaseError(3, err, "OGR Error code {}".format(err))


cdef class StackChecker:

    def __init__(self, error_stack=None):
        self.error_stack = error_stack or {}

    cdef int exc_wrap_int(self, int err) except -1:
        """Wrap a GDAL/OGR function that returns CPLErr (int).

        Raises a Rasterio exception if a non-fatal error has be set.
        """
        if err:
            stack = self.error_stack.get()
            for error, cause in zip_longest(stack[::-1], stack[::-1][1:]):
                if error is not None and cause is not None:
                    error.__cause__ = cause

            if stack:
                last = stack.pop()
                if last is not None:
                    raise last

        return err


@contextlib.contextmanager
def stack_errors():
    # TODO: better name?
    # Note: this manager produces one chain of errors and thus assumes
    # that no more than one GDAL function is called.
    CPLErrorReset()
    global _ERROR_STACK
    _ERROR_STACK.set([])

    # chaining_error_handler (better name a TODO) records GDAL errors
    # in the order they occur and converts to exceptions.
    CPLPushErrorHandlerEx(<CPLErrorHandler>chaining_error_handler, NULL)

    # Run code in the `with` block.
    yield StackChecker(_ERROR_STACK)

    CPLPopErrorHandler()
    _ERROR_STACK.set([])
    CPLErrorReset()


cdef void *exc_wrap_pointer(void *ptr) except NULL:
    """Wrap a GDAL/OGR function that returns GDALDatasetH etc (void *)

    Raises a Rasterio exception if a non-fatal error has be set.
    """
    if ptr == NULL:
        exc = exc_check()
        if exc:
            raise exc
        raise SystemError(
            f"Unknown GDAL Error. To debug: {_GDAL_DEBUG_DOCS}"
        )
    CPLErrorReset()
    return ptr


cdef VSILFILE *exc_wrap_vsilfile(VSILFILE *vsifile) except NULL:
    """Wrap a GDAL/OGR function that returns GDALDatasetH etc (void *)

    Raises a Rasterio exception if a non-fatal error has be set.

    """
    if vsifile == NULL:
        exc = exc_check()
        if exc:
            raise exc
        raise SystemError(
            f"Unknown GDAL Error. To debug: {_GDAL_DEBUG_DOCS}"
        )
    CPLErrorReset()
    return vsifile



cdef extern from "gdal.h" nogil:
    int GDALCheckVersion(int nVersionMajor, int nVersionMinor,
                         const char *pszCallingComponentName)
    const char* GDALVersionInfo(const char *pszRequest)

cdef extern from "ogr_srs_api.h" nogil:
    void OSRGetPROJVersion(int *pnMajor, int *pnMinor, int *pnPatch)

cdef extern from "ogr_core.h" nogil:
    bint OGRGetGEOSVersion(int *pnMajor, int *pnMinor, int *pnPatch)



"""Tests of overview counting and creation."""

import numpy as np
import pytest

import rasterio
from rasterio.enums import OverviewResampling
from rasterio.enums import Resampling
from rasterio.errors import OverviewCreationError


def test_count_overviews_zero(data):
    inputfile = str(data.join('RGB.byte.tif'))
    with rasterio.open(inputfile) as src:
        assert src.overviews(1) == []
        assert src.overviews(2) == []
        assert src.overviews(3) == []


def test_build_overviews_one(data):
    inputfile = str(data.join('RGB.byte.tif'))
    with rasterio.open(inputfile, 'r+') as src:
        overview_factors = [2]
        src.build_overviews(overview_factors, resampling=OverviewResampling.nearest)
        assert src.overviews(1) == [2]
        assert src.overviews(2) == [2]
        assert src.overviews(3) == [2]


def test_build_overviews_two(data):
    inputfile = str(data.join('RGB.byte.tif'))
    with rasterio.open(inputfile, 'r+') as src:
        overview_factors = [2, 4]
        src.build_overviews(overview_factors, resampling=OverviewResampling.nearest)
        assert src.overviews(1) == [2, 4]
        assert src.overviews(2) == [2, 4]
        assert src.overviews(3) == [2, 4]


def test_build_overviews_bilinear(data):
    inputfile = str(data.join('RGB.byte.tif'))
    with rasterio.open(inputfile, 'r+') as src:
        overview_factors = [2, 4]
        src.build_overviews(overview_factors, resampling=OverviewResampling.bilinear)
        assert src.overviews(1) == [2, 4]
        assert src.overviews(2) == [2, 4]
        assert src.overviews(3) == [2, 4]


def test_build_overviews_average(data):
    inputfile = str(data.join('RGB.byte.tif'))
    with rasterio.open(inputfile, 'r+') as src:
        overview_factors = [2, 4]
        src.build_overviews(overview_factors, resampling=OverviewResampling.average)
        assert src.overviews(1) == [2, 4]
        assert src.overviews(2) == [2, 4]
        assert src.overviews(3) == [2, 4]


def test_build_overviews_gauss(data):
    inputfile = str(data.join('RGB.byte.tif'))
    with rasterio.open(inputfile, 'r+') as src:
        overview_factors = [2, 4]
        src.build_overviews(overview_factors, resampling=OverviewResampling.gauss)
        assert src.overviews(1) == [2, 4]
        assert src.overviews(2) == [2, 4]
        assert src.overviews(3) == [2, 4]


def test_test_unsupported_algo(data):
    inputfile = str(data.join('RGB.byte.tif'))
    with pytest.raises(ValueError):
        with rasterio.open(inputfile, 'r+') as src:
            overview_factors = [2, 4]
            src.build_overviews(overview_factors, resampling=Resampling.q1)


def test_issue1333(data):
    """Fail if asked to create more than one 1x1 overview"""
    inputfile = str(data.join('RGB.byte.tif'))
    with pytest.raises(OverviewCreationError):
        with rasterio.open(inputfile, 'r+') as src:
            overview_factors = [1024, 2048]
            src.build_overviews(
                overview_factors, resampling=OverviewResampling.average)


def test_build_overviews_new_file(tmpdir, path_rgb_byte_tif):
    """Confirm fix of #1497"""
    dst_file = str(tmpdir.join('test.tif'))
    with rasterio.open(path_rgb_byte_tif) as src:
        with rasterio.open(dst_file, 'w', **src.profile) as dst:
            dst.write(src.read())
            overview_factors = [2, 4]
            dst.build_overviews(
                overview_factors, resampling=OverviewResampling.nearest)

    with rasterio.open(dst_file, overview_level=1) as src:
        data = src.read()
        assert data.any()


@pytest.mark.parametrize("ovr_levels", [[2], [3], [2, 4, 8]])
def test_ignore_overviews(data, ovr_levels):
    """open dataset with OVERVIEW_LEVEL=NONE, overviews should be ignored"""
    inputfile = str(data.join('RGB.byte.tif'))

    # Add overview levels to the fixture.
    with rasterio.open(inputfile, 'r+') as src:
        src.build_overviews(ovr_levels, resampling=Resampling.nearest)

    with rasterio.open(inputfile, OVERVIEW_LEVEL=-1) as src:
        assert src.overviews(1) == []
        assert src.overviews(2) == []
        assert src.overviews(3) == []


def test_decimated_no_use_overview(red_green):
    """Force ignore existing overviews when performing decimated read"""
    # Corrupt overview of red file by replacing red.tif.ovr with
    # green.tif.ovr.  We have a GDAL overview reading bug if green
    # pixels appear in a decimated read.
    green_ovr = red_green.join("green.tif.ovr")
    green_ovr.move(red_green.join("red.tif.ovr"))
    assert not green_ovr.exists()

    # Read the corrupted red overview.
    with rasterio.open(str(red_green.join("red.tif.ovr"))) as ovr:
        ovr_data = ovr.read(2)
        ovr_shape = ovr_data.shape
        assert (ovr_data == 204).all()  # Green pixels in band 2

    # Perform decimated read and ensure no use of file overview
    # (different from corrupted).
    with rasterio.open(str(red_green.join("red.tif")), OVERVIEW_LEVEL="NONE") as src:
        decimated_data = src.read(2, out_shape=ovr_shape)
        assert not np.array_equal(ovr_data, decimated_data)


def test_build_overviews_rms(data):
    """Make sure RMS resampling works"""
    inputfile = str(data.join('RGB.byte.tif'))
    with rasterio.open(inputfile, 'r+') as src:
        overview_factors = [2, 4]
        src.build_overviews(overview_factors, resampling=OverviewResampling.rms)
        assert src.overviews(1) == [2, 4]
        assert src.overviews(2) == [2, 4]
        assert src.overviews(3) == [2, 4]



# distutils: language=c++

"""Raster and vector warping and reprojection."""

from collections import UserDict
from collections.abc import Mapping
from contextlib import ExitStack
import logging
import uuid
import warnings
import xml.etree.ElementTree as ET

from affine import Affine, identity
import numpy as np

import rasterio
from rasterio._base import _transform
from rasterio._base cimport open_dataset
from rasterio._err import (
    CPLE_BaseError, CPLE_IllegalArgError, CPLE_NotSupportedError,
    CPLE_AppDefinedError, CPLE_OpenFailedError, stack_errors)
from rasterio import dtypes
from rasterio.control import GroundControlPoint
from rasterio.enums import Resampling, MaskFlags, ColorInterp
from rasterio.env import Env, GDALVersion
from rasterio.crs import CRS
from rasterio.errors import (
    GDALOptionNotImplementedError,
    DriverRegistrationError, CRSError, RasterioIOError,
    RasterioDeprecationWarning, WarpOptionsError, WarpedVRTError,
    WarpOperationError)
from rasterio.transform import Affine, from_bounds, guard_transform, tastes_like_gdal

cimport cython
cimport numpy as np
from libc.math cimport HUGE_VAL

from rasterio._base cimport get_driver_name
from rasterio._err cimport exc_wrap, exc_wrap_pointer, exc_wrap_int, StackChecker
from rasterio._io cimport (
    DatasetReaderBase, MemoryDataset, in_dtype_range, io_auto, io_band, io_multi_band)
from rasterio._features cimport GeomBuilder, OGRGeomBuilder
from rasterio.crs cimport CRS

np.import_array()
log = logging.getLogger(__name__)

# Gauss (7) is not supported for warp
SUPPORTED_RESAMPLING = [r for r in Resampling if r.value != 7]


def recursive_round(val, precision):
    """Recursively round coordinates."""
    if isinstance(val, (int, float)):
        return round(val, precision)
    else:
        return [recursive_round(part, precision) for part in val]


cdef dict TRANSFORMER_OPTIONS = None
IF (CTE_GDAL_MAJOR_VERSION, CTE_GDAL_MINOR_VERSION) >= (3, 11):
    TRANSFORMER_OPTIONS = {
        option.get("name"): option.get("description") for option
        in ET.fromstring(
            GDALGetGenImgProjTranformerOptionList().decode("utf-8")
        ).findall("Option")
    }


cdef bint is_transformer_option(bytes key):
    if TRANSFORMER_OPTIONS is None:
        return True
    return key.decode("utf-8") in TRANSFORMER_OPTIONS


cdef object _transform_single_geom(
    object single_geom,
    OGRGeometryFactory *factory,
    void *transform,
    char **options,
    int precision
):
    cdef OGRGeometryH src_geom = NULL
    cdef OGRGeometryH dst_geom = NULL

    try:
        src_geom = exc_wrap_pointer(OGRGeomBuilder().build(single_geom))
        dst_geom = exc_wrap_pointer(
            factory.transformWithOptions(
                <const OGRGeometry *>src_geom,
                <OGRCoordinateTransformation *>transform,
                options))
    else:
        result = GeomBuilder().build(dst_geom)
    finally:
        OGR_G_DestroyGeometry(dst_geom)
        OGR_G_DestroyGeometry(src_geom)

    if precision >= 0:
        # TODO: Geometry collections.
        result['coordinates'] = recursive_round(result['coordinates'], precision)

    return result


def _transform_geom(
    src_crs,
    dst_crs,
    geom,
    int precision,
):
    """Return a transformed geometry."""
    cdef char **options = NULL
    cdef OGRCoordinateTransformationH transform = NULL
    cdef OGRGeometryFactory *factory = NULL

    cdef CRS src = CRS.from_user_input(src_crs)
    cdef CRS dst = CRS.from_user_input(dst_crs)

    transform = exc_wrap_pointer(OCTNewCoordinateTransformation(src._osr, dst._osr))

    factory = new OGRGeometryFactory()

    try:
        if isinstance(geom, (dict, Mapping, UserDict)) or hasattr(geom, "__geo_interface__"):
            out_geom = _transform_single_geom(geom, factory, transform, options, precision)
        else:
            out_geom = [
                _transform_single_geom(single_geom, factory, transform, options, precision)
                for single_geom in geom
            ]
    finally:
        del factory
        OCTDestroyCoordinateTransformation(transform)
        if options != NULL:
            CSLDestroy(options)

    return out_geom


cdef GDALWarpOptions * create_warp_options(
        GDALResampleAlg resampling, object src_nodata, object dst_nodata, int src_count,
        object dst_alpha, object src_alpha, int warp_mem_limit, GDALDataType working_data_type, const char **options) except NULL:
    """Return a pointer to a GDALWarpOptions composed from input params

    This is used in _reproject() and the WarpedVRT constructor. It sets
    up warp options in almost exactly the same way as gdawarp.

    Parameters
    ----------
    dst_alpha : int
        This parameter specifies a destination alpha band for the
        warper.
    src_alpha : int
        This parameter specifies the source alpha band for the warper.

    """

    cdef GDALWarpOptions *psWOptions = GDALCreateWarpOptions()

    # Note: warp_extras is pointed to different memory locations on every
    # call to CSLSetNameValue call below, but needs to be set here to
    # get the defaults.
    cdef char **warp_extras = psWOptions.papszWarpOptions

    # See https://gdal.org/doxygen/structGDALWarpOptions.html#a0ed77f9917bb96c7a9aabd73d4d06e08
    # for a list of supported options. Copying unsupported options
    # is fine.

    # Use the same default nodata logic as gdalwarp.
    warp_extras = CSLSetNameValue(
        warp_extras, "UNIFIED_SRC_NODATA", "YES")

    warp_extras = CSLMerge(warp_extras, <char **>options)

    psWOptions.eWorkingDataType = <GDALDataType>working_data_type
    psWOptions.eResampleAlg = <GDALResampleAlg>resampling

    if warp_mem_limit > 0:
        psWOptions.dfWarpMemoryLimit = <double>warp_mem_limit * 1024 * 1024
        log.debug("Warp Memory Limit set: {!r}".format(warp_mem_limit))

    band_count = src_count

    if src_alpha:
        psWOptions.nSrcAlphaBand = src_alpha

    if dst_alpha:
        psWOptions.nDstAlphaBand = dst_alpha

    # Assign nodata values.
    # We don't currently support an imaginary component.

    if src_nodata is not None:
        psWOptions.padfSrcNoDataReal = <double*>CPLMalloc(band_count * sizeof(double))
        psWOptions.padfSrcNoDataImag = <double*>CPLMalloc(band_count * sizeof(double))

        for i in range(band_count):
            psWOptions.padfSrcNoDataReal[i] = float(src_nodata)
            psWOptions.padfSrcNoDataImag[i] = 0.0

    if dst_nodata is not None:
        psWOptions.padfDstNoDataReal = <double*>CPLMalloc(band_count * sizeof(double))
        psWOptions.padfDstNoDataImag = <double*>CPLMalloc(band_count * sizeof(double))

        for i in range(band_count):
            psWOptions.padfDstNoDataReal[i] = float(dst_nodata)
            psWOptions.padfDstNoDataImag[i] = 0.0


    # Important: set back into struct or values set above are lost
    # This is because CSLSetNameValue returns a new list each time
    psWOptions.papszWarpOptions = warp_extras

    # Set up band info
    if psWOptions.nBandCount == 0:
        psWOptions.nBandCount = band_count

        psWOptions.panSrcBands = <int*>CPLMalloc(band_count * sizeof(int))
        psWOptions.panDstBands = <int*>CPLMalloc(band_count * sizeof(int))

        for i in range(band_count):
            psWOptions.panSrcBands[i] = i + 1
            psWOptions.panDstBands[i] = i + 1

    return psWOptions


def _reproject(
        source, destination,
        src_transform=None,
        gcps=None,
        rpcs=None,
        src_crs=None,
        src_nodata=None,
        dst_transform=None,
        dst_crs=None,
        dst_nodata=None,
        dst_alpha=0,
        src_alpha=0,
        resampling=Resampling.nearest,
        init_dest_nodata=True,
        tolerance=0.125,
        num_threads=1,
        warp_mem_limit=0,
        working_data_type=0,
        src_geoloc_array=None,
        **kwargs):
    """
    Reproject a source raster to a destination raster.

    If the source and destination are ndarrays, coordinate reference
    system definitions and geolocation parameters are required for
    reprojection. Only one of src_transform, gcps, rpcs, or
    src_geoloc_array can be used.

    If the source and destination are rasterio Bands, shorthand for
    bands of datasets on disk, the coordinate reference systems and
    transforms will be read from the appropriate datasets.

    Parameters
    ------------
    source : ndarray or rasterio Band
        Source raster.
    destination : ndarray or rasterio Band
        Target raster.
    src_transform : affine.Affine(), optional
        Source affine transformation.  Required if source and destination
        are ndarrays.  Will be derived from source if it is a rasterio Band.
    gcps : sequence of `GroundControlPoint` instances, optional
        Ground control points for the source. May be used in place of
        src_transform.
    rpcs : RPC or dict, optional
        Rational polynomial coefficients for the source. May be used
        in place of src_transform.
    src_geoloc_array : array-like, optional
        A pair of 2D arrays holding x and y coordinates, like a like
        a dense array of ground control points that may be used in place
        of src_transform.
    src_crs : dict, optional
        Source coordinate reference system, in rasterio dict format.
        Required if source and destination are ndarrays.  Will be
        derived from source if it is a rasterio Band.  Example:
        'EPSG:4326'.
    src_nodata : int or float, optional
        The source nodata value.  Pixels with this value will not be used
        for interpolation.  If not set, it will be default to the
        nodata value of the source image if a masked ndarray or rasterio band,
        if available.
    dst_transform : affine.Affine(), optional
        Target affine transformation.  Required if source and destination
        are ndarrays.  Will be derived from target if it is a rasterio Band.
    dst_crs : dict, optional
        Target coordinate reference system.  Required if source and destination
        are ndarrays.  Will be derived from target if it is a rasterio Band.
    dst_nodata : int or float, optional
        The nodata value used to initialize the destination; it will remain
        in all areas not covered by the reprojected source.  Defaults to the
        nodata value of the destination image (if set), the value of
        src_nodata, or 0 (gdal default).
    src_alpha : int, optional
        Index of a band to use as the source alpha band when warping.
    dst_alpha : int, optional
        Index of a band to use as the destination alpha band when warping.
    resampling : int
        Resampling method to use.  One of the following:
            Resampling.nearest,
            Resampling.bilinear,
            Resampling.cubic,
            Resampling.cubic_spline,
            Resampling.lanczos,
            Resampling.average,
            Resampling.mode
    init_dest_nodata : bool
        Flag to specify initialization of nodata in destination;
        prevents overwrite of previous warps. Defaults to True.
    tolerance : float, optional
        The maximum error tolerance in input pixels when
        approximating the warp transformation. Default: 0.125,
        or one-eigth of a pixel.
    num_threads : int
        Number of worker threads.
    warp_mem_limit : int, optional
        The warp operation memory limit in MB. Larger values allow the
        warp operation to be carried out in fewer chunks. The amount of
        memory required to warp a 3-band uint8 2000 row x 2000 col
        raster to a destination of the same size is approximately
        56 MB. The default (0) means 64 MB with GDAL 2.2.
        The warp operation's memory limit in MB. The default (0)
        means 64 MB with GDAL 2.2.
    kwargs :  dict, optional
        Additional arguments passed to both the image to image
        transformer GDALCreateGenImgProjTransformer2() (for example,
        MAX_GCP_ORDER=2) and to the Warper (for example,
        INIT_DEST=NO_DATA).

    Returns
    ---------
    out : None
        Output is written to destination.
    """
    cdef int src_count
    cdef GDALDatasetH src_dataset = NULL
    cdef GDALDatasetH dst_dataset = NULL
    cdef char **warp_extras = NULL
    cdef const char* pszWarpThread = NULL
    cdef int i
    cdef void *hTransformArg = NULL
    cdef GDALTransformerFunc pfnTransformer = NULL
    cdef GDALWarpOptions *psWOptions = NULL
    cdef bint bUseApproxTransformer = True

    # Source geolocation parameters are mutually exclusive.
    if sum(param is not None for param in (src_transform, gcps, rpcs, src_geoloc_array)) > 1:
        raise ValueError("Source geolocation parameters are mutually exclusive.")

    # Validate nodata values immediately.
    if src_nodata is not None:
        if not in_dtype_range(src_nodata, source.dtype):
            raise ValueError("src_nodata must be in valid range for "
                             "source dtype")

    if dst_nodata is not None:
        if not in_dtype_range(dst_nodata, destination.dtype):
            raise ValueError("dst_nodata must be in valid range for "
                             "destination dtype")

    def format_transform(in_transform):
        if not in_transform:
            return in_transform
        in_transform = guard_transform(in_transform)
        # If working with identity transform, assume it is crs-less data
        # and that translating the matrix very slightly will avoid #674 and #1272
        eps = 1e-100
        if in_transform.almost_equals(identity) or in_transform.almost_equals(Affine(1, 0, 0, 0, -1, 0)):
            in_transform = in_transform.translation(eps, eps)
        return in_transform

    cdef:
        MemoryDataset mem_raster = None
        MemoryDataset src_mem = None
        char **imgProjOptions = NULL
        GDALRasterBandH hBand = NULL
        GDALWarpOperation oWarper
        int rows
        int cols
        StackChecker checker

    with ExitStack() as exit_stack:
        # If the source is an ndarray, we copy to a MEM dataset.
        # We need a src_transform and src_dst in this case. These will
        # be copied to the MEM dataset.
        if dtypes.is_ndarray(source):
            if not src_crs:
                raise CRSError("Missing src_crs.")

            # ensure data converted to numpy array
            if hasattr(source, "mask"):
                source = np.ma.asanyarray(source)
            else:
                source = np.asanyarray(source)

            # Convert 2D single-band arrays to 3D multi-band.
            if len(source.shape) == 2:
                source = source.reshape(1, *source.shape)

            src_count = source.shape[0]
            src_bidx = range(1, src_count + 1)

            if hasattr(source, "mask"):
                if source.mask is np.ma.nomask:
                    if source.ndim == 2:
                        mask_shape = source.shape
                    else:
                        mask_shape = source.shape[1:]
                    mask = np.full(mask_shape, np.uint8(255))
                else:
                    mask = ~np.logical_or.reduce(source.mask) * np.uint8(255)
                source_arr = np.concatenate((source.data, [mask]))
                src_alpha = src_alpha or source_arr.shape[0]
            else:
                source_arr = source

            src_mem = exit_stack.enter_context(
                MemoryDataset(
                    source_arr,
                    transform=format_transform(src_transform),
                    gcps=gcps,
                    rpcs=rpcs,
                    crs=src_crs,
                    copy=True,
                )
            )
            src_dataset = src_mem.handle()

            if src_alpha:
                for i in range(source_arr.shape[0]):
                    GDALDeleteRasterNoDataValue(GDALGetRasterBand(src_dataset, i+1))
                GDALSetRasterColorInterpretation(
                    GDALGetRasterBand(src_dataset, src_alpha),
                    <GDALColorInterp>6,
                )

        # If the source is a rasterio MultiBand, no copy necessary.
        # A MultiBand is a tuple: (dataset, bidx, dtype, shape(2d))
        elif isinstance(source, tuple):
            rdr, src_bidx, dtype, shape = source
            if isinstance(src_bidx, int):
                src_bidx = [src_bidx]
            src_count = len(src_bidx)
            src_dataset = (<DatasetReaderBase?>rdr).handle()
            if src_nodata is None:
                src_nodata = rdr.nodata

        else:
            raise ValueError("Invalid source")

        # Next, do the same for the destination raster.
        dest_2d = False

        if dtypes.is_ndarray(destination):
            if not dst_crs:
                raise CRSError("Missing dst_crs.")

            dst_nodata = dst_nodata or src_nodata

            if hasattr(destination, "mask"):
                destination = np.ma.asanyarray(destination)
            else:
                destination = np.asanyarray(destination)

            if len(destination.shape) == 2:
                dest_2d = True
                destination = destination.reshape(1, *destination.shape)

            if destination.shape[0] == src_count:
                # Output shape matches number of bands being extracted
                dst_bidx = [i + 1 for i in range(src_count)]
            else:
                # Assume src and dst are the same shape
                if max(src_bidx) > destination.shape[0]:
                    raise ValueError("Invalid destination shape")
                dst_bidx = src_bidx

            if hasattr(destination, "mask"):
                count, height, width = destination.shape
                msk = np.logical_or.reduce(destination.mask)
                if msk == np.ma.nomask:
                    msk = np.zeros((height, width), dtype=bool)
                msk = ~msk * np.uint8(255)
                dest_arr = np.concatenate((destination.data, [msk]))
                dst_alpha = dst_alpha or dest_arr.shape[0]
            else:
                dest_arr = destination

            mem_raster = exit_stack.enter_context(
                MemoryDataset(dest_arr, transform=format_transform(dst_transform), crs=dst_crs)
            )
            dst_dataset = mem_raster.handle()

            if dst_alpha:
                for i in range(dest_arr.shape[0]):
                    GDALDeleteRasterNoDataValue(GDALGetRasterBand(dst_dataset, i+1))
                GDALSetRasterColorInterpretation(
                    GDALGetRasterBand(dst_dataset, dst_alpha),
                    <GDALColorInterp>6,
                )

            GDALSetDescription(
                dst_dataset, "Temporary destination dataset for _reproject()")

            log.debug("Created temp destination dataset.")

        elif isinstance(destination, tuple):
            udr, dst_bidx, _, _ = destination
            if isinstance(dst_bidx, int):
                dst_bidx = [dst_bidx]
            dst_dataset = (<DatasetReaderBase?>udr).handle()
            if dst_nodata is None:
                dst_nodata = udr.nodata
        else:
            raise ValueError("Invalid destination")


        # Set up GDALCreateGenImgProjTransformer2 keyword arguments.
        imgProjOptions = CSLSetNameValue(imgProjOptions, "GCPS_OK", "TRUE")

        if rpcs:
            imgProjOptions = CSLSetNameValue(imgProjOptions, "SRC_METHOD", "RPC")
        elif src_geoloc_array is not None:
            arr = np.stack((src_geoloc_array[0], src_geoloc_array[1]))
            geoloc_dataset = exit_stack.enter_context(MemoryDataset(arr, crs=src_crs))
            log.debug("Geoloc dataset created: geoloc_dataset=%r", geoloc_dataset)
            imgProjOptions = CSLSetNameValue(
                imgProjOptions, "SRC_GEOLOC_ARRAY", geoloc_dataset.name.encode("utf-8")
            )
            imgProjOptions = CSLSetNameValue(
                imgProjOptions, "SRC_SRS", src_crs.to_string().encode("utf-8")
            )
            bUseApproxTransformer = False

        # See https://gdal.org/doxygen/gdal__alg_8h.html#a94cd172f78dbc41d6f407d662914f2e3
        # for a list of supported options. I (Sean) don't see harm in
        # copying all the function's keyword arguments to the image to
        # image transformer options mapping; unsupported options should be
        # okay.
        for key, val in kwargs.items():
            key = key.upper().encode('utf-8')

            if not is_transformer_option(key):
                continue

            if key in {b"RPC_DEM", b"COORDINATE_OPERATION"}:
                # don't .upper() since might be a path
                val = str(val).encode('utf-8')

                if rpcs:
                    bUseApproxTransformer = False
            else:
                val = str(val).upper().encode('utf-8')
            imgProjOptions = CSLSetNameValue(
                imgProjOptions, <const char *>key, <const char *>val)
            log.debug("Set _reproject Transformer option {0!r}={1!r}".format(key, val))

        try:
            with Env(GDAL_MEM_ENABLE_OPEN=True):
                hTransformArg = exc_wrap_pointer(
                    GDALCreateGenImgProjTransformer2(
                        src_dataset, dst_dataset, imgProjOptions))
            if bUseApproxTransformer:
                hTransformArg = exc_wrap_pointer(
                    GDALCreateApproxTransformer(
                        GDALGenImgProjTransform, hTransformArg, tolerance))
                pfnTransformer = GDALApproxTransform
                GDALApproxTransformerOwnsSubtransformer(hTransformArg, 1)
                log.debug("Created approximate transformer")
            else:
                pfnTransformer = GDALGenImgProjTransform
                log.debug("Created exact transformer")

            log.debug("Created transformer and options.")
        except:
            if bUseApproxTransformer:
                GDALDestroyApproxTransformer(hTransformArg)
            else:
                GDALDestroyGenImgProjTransformer(hTransformArg)
            CPLFree(imgProjOptions)
            raise

        valb = str(num_threads).encode('utf-8')
        warp_extras = CSLSetNameValue(warp_extras, "NUM_THREADS", <const char *>valb)

        log.debug("Setting NUM_THREADS option: %d", num_threads)

        if init_dest_nodata:
            warp_extras = CSLSetNameValue(warp_extras, "INIT_DEST",
                                        "NO_DATA" if dst_nodata is not None else "0")

        # See https://gdal.org/doxygen/structGDALWarpOptions.html#a0ed77f9917bb96c7a9aabd73d4d06e08
        # for a list of supported options. Copying unsupported options
        # is fine.
        for key, val in kwargs.items():
            key = key.upper().encode('utf-8')
            val = str(val).upper().encode('utf-8')
            warp_extras = CSLSetNameValue(
                warp_extras, <const char *>key, <const char *>val)

        psWOptions = create_warp_options(
            <GDALResampleAlg>resampling,
            src_nodata,
            dst_nodata,
            src_count,
            dst_alpha,
            src_alpha,
            warp_mem_limit,
            <GDALDataType>working_data_type,
            <const char **>warp_extras
        )

        psWOptions.pfnTransformer = pfnTransformer
        psWOptions.pTransformerArg = hTransformArg
        psWOptions.hSrcDS = src_dataset
        psWOptions.hDstDS = dst_dataset

        for idx, (s, d) in enumerate(zip(src_bidx, dst_bidx)):
            psWOptions.panSrcBands[idx] = s
            psWOptions.panDstBands[idx] = d
            log.debug('Configured to warp src band %d to destination band %d' % (s, d))

        log.debug("Set transformer options")

        # Now that the transformer and warp options are set up, we init
        # and run the warper.
        try:
            exc_wrap_int(oWarper.Initialize(psWOptions))
            if isinstance(destination, tuple):
                rows, cols = destination[3]
            else:
                rows, cols = destination.shape[-2:]

            log.debug(
                "Chunk and warp window: %d, %d, %d, %d.",
                0, 0, cols, rows)

            try:
                with stack_errors() as checker, Env(GDAL_MEM_ENABLE_OPEN=True):
                    if num_threads > 1:
                        with nogil:
                            err = oWarper.ChunkAndWarpMulti(0, 0, cols, rows)
                    else:
                        with nogil:
                            err = oWarper.ChunkAndWarpImage(0, 0, cols, rows)

                    err = checker.exc_wrap_int(err)

            except CPLE_BaseError as base:
                raise WarpOperationError("Chunk and warp failed") from base

            if mem_raster is not None:
                count, height, width = dest_arr.shape
                if hasattr(destination, "mask"):
                    # Pick off the alpha band and make a mask of it.
                    # TODO: do this efficiently, not copying unless necessary.
                    indexes = np.arange(1, count, dtype='intp')
                    io_multi_band(dst_dataset, 0, 0.0, 0.0, width, height, destination, indexes)
                    alpha_arr = np.empty((height, width), dtype=dest_arr.dtype)
                    io_band(mem_raster.band(count), 0, 0.0, 0.0, width, height, alpha_arr)
                    destination = np.ma.masked_array(
                        destination.data,
                        mask=np.repeat(
                            ~(alpha_arr.astype("bool"))[np.newaxis, :, :],
                            count - 1,
                            axis=0,
                        )
                    )
                else:
                    exc_wrap_int(io_auto(destination, dst_dataset, 0))

                if dest_2d:
                    destination = destination[0]

                return destination

        # Clean up transformer, warp options, and dataset handles.
        finally:
            if bUseApproxTransformer:
                GDALDestroyApproxTransformer(hTransformArg)
            else:
                GDALDestroyGenImgProjTransformer(hTransformArg)

            GDALDestroyWarpOptions(psWOptions)
            CPLFree(imgProjOptions)
            CSLDestroy(warp_extras)


def _calculate_default_transform(
    src_crs,
    dst_crs,
    width,
    height,
    left=None,
    bottom=None,
    right=None,
    top=None,
    gcps=None,
    rpcs=None,
    src_geoloc_array=None,
    **kwargs
):
    """Wraps GDAL's algorithm."""
    cdef void *hTransformArg = NULL
    cdef int npixels = 0
    cdef int nlines = 0
    cdef double extent[4]
    cdef double geotransform[6]
    cdef char *wkt = NULL
    cdef GDALDatasetH hds = NULL
    cdef char **imgProjOptions = NULL
    cdef char **papszMD = NULL
    cdef bint bUseApproxTransformer = True

    # Source geolocation parameters are mutually exclusive.
    # if sum(param is not None for param in (left, gcps, rpcs, src_geoloc_array)) > 1:
    #     raise ValueError("Source geolocation parameters are mutually exclusive.")

    extent[:] = [0.0, 0.0, 0.0, 0.0]
    geotransform[:] = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    if all(x is not None for x in (left, bottom, right, top)):
        transform = from_bounds(left, bottom, right, top, width, height)
    elif any(x is not None for x in (left, bottom, right, top)):
        raise ValueError(
            "Some, but not all, bounding box parameters were provided.")
    else:
        transform = None

    dst_crs = CRS.from_user_input(dst_crs)
    wkt_b = dst_crs.to_wkt().encode('utf-8')
    wkt = wkt_b

    if src_crs is not None:
        src_crs = CRS.from_user_input(src_crs)

    # The transformer at the heart of this function requires a dataset.
    # We use an in-memory VRT dataset.
    vrt_doc = _suggested_proxy_vrt_doc(
        width,
        height,
        transform=transform,
        crs=src_crs,
        gcps=gcps,
        rpcs=rpcs,
    ).decode('utf-8')

    hds = open_dataset(vrt_doc, 0x00 | 0x02 | 0x04, ['VRT'], {}, None)

    with ExitStack() as exit_stack:
        try:
            imgProjOptions = CSLSetNameValue(imgProjOptions, "GCPS_OK", "TRUE")
            imgProjOptions = CSLSetNameValue(imgProjOptions, "MAX_GCP_ORDER", "0")
            imgProjOptions = CSLSetNameValue(imgProjOptions, "DST_SRS", wkt)

            for key, val in kwargs.items():
                key = key.upper().encode('utf-8')

                if not is_transformer_option(key):
                    continue

                if key in {b"RPC_DEM", b"COORDINATE_OPERATION"}:
                    # don't .upper() since might be a path.
                    val = str(val).encode('utf-8')
                else:
                    val = str(val).upper().encode('utf-8')

                imgProjOptions = CSLSetNameValue(imgProjOptions, <const char *>key, <const char *>val)
                log.debug("Set image projection option {0!r}={1!r}".format(key, val))

            if rpcs:
                if hasattr(rpcs, 'to_gdal'):
                    rpcs = rpcs.to_gdal()

                for key, val in rpcs.items():
                    key = key.upper().encode('utf-8')
                    val = str(val).encode('utf-8')
                    papszMD = CSLSetNameValue(papszMD, <const char *>key, <const char *>val)

                exc_wrap_int(GDALSetMetadata(hds, papszMD, "RPC"))
                imgProjOptions = CSLSetNameValue(imgProjOptions, "SRC_METHOD", "RPC")
                bUseApproxTransformer = False

            elif src_geoloc_array is not None:
                arr = np.stack((src_geoloc_array[0], src_geoloc_array[1]))
                geoloc_dataset = exit_stack.enter_context(MemoryDataset(arr, crs=src_crs))
                log.debug("Geoloc dataset created: geoloc_dataset=%r", geoloc_dataset)
                imgProjOptions = CSLSetNameValue(
                    imgProjOptions, "SRC_GEOLOC_ARRAY", geoloc_dataset.name.encode("utf-8")
                )
                imgProjOptions = CSLSetNameValue(
                    imgProjOptions, "SRC_SRS", src_crs.to_string().encode("utf-8")
                )
                bUseApproxTransformer = False

            with Env(GDAL_MEM_ENABLE_OPEN=True):
                hTransformArg = exc_wrap_pointer(
                    GDALCreateGenImgProjTransformer2(hds, NULL, imgProjOptions)
                )

            pfnTransformer = GDALGenImgProjTransform
            log.debug("Created exact transformer")

            exc_wrap_int(
                GDALSuggestedWarpOutput2(
                    hds,
                    pfnTransformer,
                    hTransformArg,
                    geotransform,
                    &npixels,
                    &nlines,
                    extent,
                    0
                )
            )

        except CPLE_NotSupportedError as err:
            raise CRSError(err.errmsg)

        finally:
            if hTransformArg != NULL:
                GDALDestroyGenImgProjTransformer(hTransformArg)
            if hds != NULL:
                GDALClose(hds)
            if imgProjOptions != NULL:
                CPLFree(imgProjOptions)
            if papszMD != NULL:
                CSLDestroy(papszMD)

    dst_affine = Affine.from_gdal(*[geotransform[i] for i in range(6)])
    dst_width = npixels
    dst_height = nlines

    return (dst_affine, dst_width, dst_height)


DEFAULT_NODATA_FLAG = object()

cdef GDALDatasetH auto_create_warped_vrt(
    GDALDatasetH hSrcDS,
    const char *pszSrcWKT,
    const char *pszDstWKT,
    GDALResampleAlg eResampleAlg,
    double dfMaxError,
    const GDALWarpOptions *psOptions,
    const char **transformer_options,
) except NULL:
    """Makes a best-fit WarpedVRT around a dataset.

    Returns
    -------
    GDALDatasetH
        Owned by the caller.

    """
    cdef GDALDatasetH hds_warped = NULL

    try:
        # This function may put errors on GDAL's error stack while
        # still returning 0 (no error). Thus we always check and
        # clear the stack and ignore the error if the function
        # succeeds.
        with nogil:
            hds_warped = GDALAutoCreateWarpedVRTEx(
                hSrcDS,
                pszSrcWKT,
                pszDstWKT,
                eResampleAlg,
                dfMaxError,
                psOptions,
                transformer_options,
            )
        _ = exc_wrap(1)

    except CPLE_AppDefinedError as err:
        if hds_warped != NULL:
            log.info("Ignoring error: err=%r", err)
        else:
            raise err

    return hds_warped


cdef class WarpedVRTReaderBase(DatasetReaderBase):

    def __init__(self, src_dataset, src_crs=None, crs=None,
                 resampling=Resampling.nearest, tolerance=0.125,
                 src_nodata=DEFAULT_NODATA_FLAG, nodata=DEFAULT_NODATA_FLAG,
                 width=None, height=None,
                 src_transform=None, transform=None,
                 init_dest_nodata=True, src_alpha=0, dst_alpha=0, add_alpha=False,
                 warp_mem_limit=0, dtype=None, **warp_extras):
        """Make a virtual warped dataset

        Parameters
        ----------
        src_dataset : dataset object
            The warp source dataset. Must be opened in "r" mode.
        src_crs : CRS or str, optional
            Overrides the coordinate reference system of `src_dataset`.
        src_transfrom : Affine, optional
            Overrides the transform of `src_dataset`.
        src_nodata : float, optional
            Overrides the nodata value of `src_dataset`, which is the
            default.
        crs : CRS or str, optional
            The coordinate reference system at the end of the warp
            operation.  Default: the crs of `src_dataset`. dst_crs was
            a deprecated alias for this parameter.
        transform : Affine, optional
            The transform for the virtual dataset. Default: will be
            computed from the attributes of `src_dataset`. dst_transform
            was a deprecated alias for this parameter.
        height, width: int, optional
            The dimensions of the virtual dataset. Defaults: will be
            computed from the attributes of `src_dataset`. dst_height
            and dst_width were deprecated alias for these parameters.
        nodata : float, optional
            Nodata value for the virtual dataset. Default: the nodata
            value of `src_dataset` or 0.0. dst_nodata was a deprecated
            alias for this parameter.
        resampling : Resampling, optional
            Warp resampling algorithm. Default: `Resampling.nearest`.
        tolerance : float, optional
            The maximum error tolerance in input pixels when
            approximating the warp transformation. Default: 0.125,
            or one-eigth of a pixel.
        src_alpha : int, optional
            Index of a source band to use as an alpha band for warping.
        dst_alpha : int, optional
            Index of a destination band to use as an alpha band for warping.
        add_alpha : bool, optional
            Whether to add an alpha masking band to the virtual dataset.
            Default: False. This option will cause deletion of the VRT
            nodata value.
        init_dest_nodata : bool, optional
            Whether or not to initialize output to `nodata`. Default:
            True.
        warp_mem_limit : int, optional
            The warp operation's memory limit in MB. The default (0)
            means 64 MB with GDAL 2.2.
        dtype : str, optional
            The working data type for warp operation and output.
        warp_extras : dict, optional
            GDAL extra warp options. See:
            https://gdal.org/doxygen/structGDALWarpOptions.html.
            Also, GDALCreateGenImgProjTransformer2() options.
            Requires rasterio 1.3+, GDAL 3.2+. See:
            https://gdal.org/doxygen/gdal__alg_8h.html#a94cd172f78dbc41d6f407d662914f2e3

        Returns
        -------
        WarpedVRT

        """
        if src_dataset.mode != "r":
            warnings.warn("Source dataset should be opened in read-only mode. Use of datasets opened in modes other than 'r' will be disallowed in a future version.", RasterioDeprecationWarning, stacklevel=2)

        # Guard against invalid or unsupported resampling algorithms.
        try:
            if resampling == 7:
                raise ValueError("Gauss resampling is not supported")
            Resampling(resampling)

        except ValueError:
            raise ValueError(
                "resampling must be one of: {0}".format(", ".join(['Resampling.{0}'.format(r.name) for r in SUPPORTED_RESAMPLING])))

        self.mode = 'r'
        self.options = {}
        self._count = 0
        self._dtypes = []
        self._block_shapes = None
        self._nodatavals = []
        self._units = ()
        self._descriptions = ()
        self._crs = None
        self._gcps = None
        self._read = False

        # kwargs become warp options.
        self.src_dataset = src_dataset
        self.src_crs = src_dataset.crs if src_crs is None else CRS.from_user_input(src_crs)
        self.src_transform = src_transform
        self.src_gcps = None
        self.name = "WarpedVRT({})".format(src_dataset.name)
        self.resampling = resampling
        self.tolerance = tolerance
        self.working_dtype = dtype

        self.src_nodata = self.src_dataset.nodata if src_nodata is DEFAULT_NODATA_FLAG else src_nodata
        self.dst_nodata = self.src_nodata if nodata is DEFAULT_NODATA_FLAG else nodata
        self.dst_width = width
        self.dst_height = height
        self.dst_transform = transform
        self.warp_extras = warp_extras.copy()
        if init_dest_nodata is True and 'init_dest' not in warp_extras:
            self.warp_extras['init_dest'] = 'NO_DATA' if self.dst_nodata is not None else '0'

        cdef GDALDriverH driver = NULL
        cdef GDALDatasetH hds = NULL
        cdef GDALDatasetH hds_warped = NULL
        cdef const char *cypath = NULL
        cdef char *src_crs_wkt = NULL
        cdef char *dst_crs_wkt = NULL
        cdef OGRSpatialReferenceH osr = NULL
        cdef char **c_warp_extras = NULL
        cdef char **c_transformer_options = NULL
        cdef GDALWarpOptions *psWOptions = NULL
        cdef float c_tolerance = tolerance
        cdef GDALResampleAlg c_resampling = resampling
        cdef int c_width = self.dst_width or 0
        cdef int c_height = self.dst_height or 0
        cdef double src_gt[6]
        cdef double dst_gt[6]
        cdef void *hTransformArg = NULL
        cdef GDALRasterBandH hband = NULL
        cdef GDALRasterBandH hmask = NULL
        cdef int mask_block_xsize = 0
        cdef int mask_block_ysize = 0

        hds = exc_wrap_pointer((<DatasetReaderBase?>self.src_dataset).handle())

        if not self.src_transform:
            self.src_transform = self.src_dataset.transform

        if not self.src_crs and self.src_transform.almost_equals(identity):
            try:
                self.src_gcps, self.src_crs = self.src_dataset.get_gcps()
            except ValueError:
                pass

        self.dst_crs = CRS.from_user_input(crs) if crs is not None else self.src_crs

        # Convert CRSes to C WKT strings.
        if self.src_crs is not None:
            src_crs_wkt_b = self.src_crs.to_wkt().encode("utf-8")
            src_crs_wkt = src_crs_wkt_b

        if self.dst_crs is not None:
            dst_crs_wkt_b = self.dst_crs.to_wkt().encode("utf-8")
            dst_crs_wkt = dst_crs_wkt_b

        log.debug("Exported CRS to WKT.")

        log.debug("Warp_extras: %r", self.warp_extras)

        for key, val in self.warp_extras.items():
            key = key.upper().encode('utf-8')
            val = str(val).upper().encode('utf-8')
            c_warp_extras = CSLSetNameValue(
                c_warp_extras, <const char *>key, <const char *>val)
            if is_transformer_option(key):
                c_transformer_options = CSLSetNameValue(
                    c_transformer_options, <const char *>key, <const char *>val)

        cdef GDALRasterBandH hBand = NULL
        src_alpha_band = 0
        dst_alpha_band = 0
        for bidx in src_dataset.indexes:
            hBand = GDALGetRasterBand(hds, bidx)
            if GDALGetRasterColorInterpretation(hBand) == GCI_AlphaBand:
                src_alpha_band = bidx
                dst_alpha_band = bidx

        # Adding an alpha band when the source has one is trouble.
        # It will result in surprisingly unmasked data. We will
        # raise an exception instead.

        if add_alpha:
            if src_alpha_band:
                raise WarpOptionsError(
                    "The VRT already has an alpha band, adding a new one is not supported")
            else:
                dst_alpha_band = src_dataset.count + 1
                self.dst_nodata = None

        if dst_alpha:
            dst_alpha_band = dst_alpha

        if src_alpha:
            src_alpha_band = src_alpha

        gdal_dtype = dtypes._get_gdal_dtype(self.working_dtype)
        psWOptions = create_warp_options(
            <GDALResampleAlg>c_resampling, self.src_nodata,
            self.dst_nodata, src_dataset.count, dst_alpha_band,
            src_alpha_band, warp_mem_limit, <GDALDataType>gdal_dtype,
            <const char **>c_warp_extras)

        if psWOptions == NULL:
            raise RuntimeError("Warp options are NULL")

        psWOptions.hSrcDS = hds

        # We handle four different use cases.
        #
        # 1. Destination transform, height, and width are provided by
        #    the caller.
        # 2. Pure scaling: source and destination CRS are the same,
        #    destination height and width are provided by the caller,
        #    destination transform is not provided; it is computed by
        #    scaling the source transform.
        # 3. Warp with scaling: CRS are different, destination height
        #    and width are provided by the caller, destination transform
        #    is computed and scaled to preserve given dimensions.
        # 4. Warp with destination height, width, and transform
        #    auto-generated by GDAL.

        # Case 4
        if not self.dst_transform and (not self.dst_width or not self.dst_height):
            try:
                hds_warped = auto_create_warped_vrt(
                    hds,
                    src_crs_wkt,
                    dst_crs_wkt,
                    c_resampling,
                    c_tolerance,
                    psWOptions,
                    <const char **>c_transformer_options,
                )
            finally:
                CSLDestroy(c_warp_extras)
                CSLDestroy(c_transformer_options)
                if psWOptions != NULL:
                    GDALDestroyWarpOptions(psWOptions)

        else:
            # Case 1
            if self.dst_transform and self.dst_width and self.dst_height:
                pass

            # Case 2
            elif (
                self.src_crs == self.dst_crs
                and self.dst_width
                and self.dst_height
                and not self.dst_transform
                and not self.src_gcps
                and not self.src_dataset.rpcs
            ):
                # Note: scaling on the right hand side of multiplication
                # preserves the origin of the geotransform matrix.
                self.dst_transform = self.src_transform * Affine.scale(
                    self.src_dataset.width / self.dst_width,
                    self.src_dataset.height / self.dst_height
                )

            # Case 3
            elif (
                (self.src_crs != self.dst_crs or self.src_gcps or self.src_dataset.rpcs)
                and self.dst_width
                and self.dst_height
                and not self.dst_transform
            ):
                left, bottom, right, top = self.src_dataset.bounds
                self.dst_transform, width, height = _calculate_default_transform(
                    self.src_crs,
                    self.dst_crs,
                    self.src_dataset.width,
                    self.src_dataset.height,
                    left=left,
                    bottom=bottom,
                    right=right,
                    top=top,
                    gcps=self.src_gcps,
                    rpcs=self.src_dataset.rpcs,
                    **self.warp_extras,
                )
                self.dst_transform = self.dst_transform * Affine.scale(
                    width / self.dst_width, height / self.dst_height
                )

            # If we get here it's because the tests above are buggy.
            # We raise a Python exception to indicate that.
            else:
                CSLDestroy(c_warp_extras)
                CSLDestroy(c_transformer_options)
                if psWOptions != NULL:
                    GDALDestroyWarpOptions(psWOptions)
                raise RuntimeError("Parameterization error")

            # Continue with finishing cases 1-3, which have been
            # generalized.
            t = self.src_transform.to_gdal()
            for i in range(6):
                src_gt[i] = t[i]

            t = self.dst_transform.to_gdal()
            for i in range(6):
                dst_gt[i] = t[i]

            try:
                hTransformArg = exc_wrap_pointer(
                    GDALCreateGenImgProjTransformer3(
                        src_crs_wkt, src_gt, dst_crs_wkt, dst_gt))

                if c_tolerance > 0.0:

                    hTransformArg = exc_wrap_pointer(
                        GDALCreateApproxTransformer(
                            GDALGenImgProjTransform,
                            hTransformArg,
                            c_tolerance))

                    psWOptions.pfnTransformer = GDALApproxTransform
                    GDALApproxTransformerOwnsSubtransformer(hTransformArg, 1)

                psWOptions.pTransformerArg = hTransformArg

                with nogil:
                    hds_warped = GDALCreateWarpedVRT(hds, c_width, c_height, dst_gt, psWOptions)
                    GDALSetProjection(hds_warped, dst_crs_wkt)

            finally:
                CSLDestroy(c_warp_extras)
                CSLDestroy(c_transformer_options)
                if psWOptions != NULL:
                    GDALDestroyWarpOptions(psWOptions)

        # End of the 4 cases.

        try:
            self._hds = exc_wrap_pointer(hds_warped)

        except CPLE_OpenFailedError as err:
            raise RasterioIOError(err.errmsg)

        if self.dst_nodata is None:
            for i in self.indexes:
                GDALDeleteRasterNoDataValue(self.band(i))

        else:
            for i in self.indexes:
                GDALSetRasterNoDataValue(self.band(i), self.dst_nodata)

        if dst_alpha:
            GDALSetRasterColorInterpretation(self.band(dst_alpha), <GDALColorInterp>6)

        self._set_attrs_from_dataset_handle()
        self._nodatavals = [self.dst_nodata for i in self.indexes]

        if dst_alpha and len(self._nodatavals) == 3:
            self._nodatavals[dst_alpha - 1] = None

        self._env = ExitStack()

    @property
    def crs(self):
        """The dataset's coordinate reference system"""
        return self.dst_crs

    def read(self, indexes=None, out=None, window=None, masked=False, out_shape=None, resampling=Resampling.nearest, fill_value=None, out_dtype=None, **kwargs):
        """Read a dataset's raw pixels as an N-d array

        This data is read from the dataset's band cache, which means
        that repeated reads of the same windows may avoid I/O.

        Parameters
        ----------
        indexes : list of ints or a single int, optional
            If `indexes` is a list, the result is a 3D array, but is
            a 2D array if it is a band index number.

        out : numpy.ndarray, optional
            As with Numpy ufuncs, this is an optional reference to an
            output array into which data will be placed. If the height
            and width of `out` differ from that of the specified
            window (see below), the raster image will be decimated or
            replicated using the specified resampling method (also see
            below).

            *Note*: the method's return value may be a view on this
            array. In other words, `out` is likely to be an
            incomplete representation of the method's results.

            This parameter cannot be combined with `out_shape`.

        out_dtype : str or numpy.dtype
            The desired output data type. For example: 'uint8' or
            rasterio.uint16.

        out_shape : tuple, optional
            A tuple describing the shape of a new output array. See
            `out` (above) for notes on image decimation and
            replication.

            Cannot combined with `out`.

        window : a pair (tuple) of pairs of ints or Window, optional
            The optional `window` argument is a 2 item tuple. The first
            item is a tuple containing the indexes of the rows at which
            the window starts and stops and the second is a tuple
            containing the indexes of the columns at which the window
            starts and stops. For example, ((0, 2), (0, 2)) defines
            a 2x2 window at the upper left of the raster dataset.

        masked : bool, optional
            If `masked` is `True` the return value will be a masked
            array. Otherwise (the default) the return value will be a
            regular array. Masks will be exactly the inverse of the
            GDAL RFC 15 conforming arrays returned by read_masks().

        resampling : Resampling
            By default, pixel values are read raw or interpolated using
            a nearest neighbor algorithm from the band cache. Other
            resampling algorithms may be specified. Resampled pixels
            are not cached.

        fill_value : scalar
            Fill value applied in the `boundless=True` case only.

        kwargs : dict
            This is only for backwards compatibility. No keyword arguments
            are supported other than the ones named above.

        Returns
        -------
        numpy.ndarray or a view on a numpy.ndarray

        Note: as with Numpy ufuncs, an object is returned even if you
        use the optional `out` argument and the return value shall be
        preferentially used by callers.

        """
        if kwargs.get("boundless", False):
            raise ValueError("WarpedVRT does not permit boundless reads")
        else:
            return super().read(indexes=indexes, out=out, window=window, masked=masked, out_shape=out_shape, resampling=resampling, fill_value=fill_value, out_dtype=out_dtype)

    def read_masks(self, indexes=None, out=None, out_shape=None, window=None, resampling=Resampling.nearest, **kwargs):
        """Read raster band masks as a multidimensional array"""
        if kwargs.get("boundless", False):
            raise ValueError("WarpedVRT does not permit boundless reads")
        else:
            return super().read_masks(indexes=indexes, out=out, window=window, out_shape=out_shape, resampling=resampling)


def _suggested_proxy_vrt_doc(width, height, transform=None, crs=None, gcps=None, rpcs=None):
    """Make a VRT XML document to serve _calculate_default_transform."""
    vrtdataset = ET.Element('VRTDataset')
    vrtdataset.attrib['rasterYSize'] = str(height)
    vrtdataset.attrib['rasterXSize'] = str(width)
    vrtrasterband = ET.SubElement(vrtdataset, 'VRTRasterBand')

    srs = ET.SubElement(vrtdataset, 'SRS')
    srs.text = crs.wkt if crs else ""

    if gcps:
        gcplist = ET.SubElement(vrtdataset, 'GCPList')
        gcplist.attrib['Projection'] = crs.wkt if crs else ""
        for point in gcps:
            gcp = ET.SubElement(gcplist, 'GCP')
            gcp.attrib['Id'] = str(point.id)
            gcp.attrib['Info'] = str(point.info)
            gcp.attrib['Pixel'] = str(point.col)
            gcp.attrib['Line'] = str(point.row)
            gcp.attrib['X'] = str(point.x)
            gcp.attrib['Y'] = str(point.y)
            gcp.attrib['Z'] = str(point.z)
    elif transform:
        geotransform = ET.SubElement(vrtdataset, 'GeoTransform')
        geotransform.text = ','.join([str(v) for v in transform.to_gdal()])
    elif rpcs:
        metadata = ET.SubElement(vrtdataset, "Metadata")
        metadata.attrib["domain"] = "RPC"
        for key in (
            "line_off", "samp_off", "height_off", "lat_off", "long_off",
            "line_scale", "samp_scale", "height_scale", "lat_scale", "long_scale",
            "err_bias", "err_rand",
        ):
            item = ET.SubElement(metadata, "MDI")
            item.attrib["key"] = key.upper()
            item.text = str(getattr(rpcs, key))
        for key in ("line_num_coeff", "line_den_coeff", "samp_num_coeff", "samp_den_coeff"):
            item = ET.SubElement(metadata, "MDI")
            item.attrib["key"] = key.upper()
            item.text = " ".join(["{:+e}".format(val) for val in getattr(rpcs, key)])

    return ET.tostring(vrtdataset)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef double antimeridian_min(data):
    """
    Handles the case when longitude values cross the antimeridian
    when calculating the minimum.

    Note: The data array must be in a linear ring.

    Note: This requires a densified ring with at least 2 additional
          points per edge to correctly handle global extents.

    If only 1 additional point:

      |        |
      |RL--x0--|RL--
      |        |
    -180    180|-180

    If they are evenly spaced and it crosses the antimeridian:
    x0 - L = 180
    R - x0 = -180

    For example:
    Let R = -179.9, x0 = 0.1, L = -179.89
    x0 - L = 0.1 - -179.9 = 180
    R - x0 = -179.89 - 0.1 ~= -180
    This is the same in the case when it didn't cross the antimeridian.

    If you have 2 additional points:
      |            |
      |RL--x0--x1--|RL--
      |            |
    -180        180|-180

    If they are evenly spaced and it crosses the antimeridian:
    x0 - L = 120
    x1 - x0 = 120
    R - x1 = -240

    For example:
    Let R = -179.9, x0 = -59.9, x1 = 60.1 L = -179.89
    x0 - L = 59.9 - -179.9 = 120
    x1 - x0 = 60.1 - 59.9 = 120
    R - x1 = -179.89 - 60.1 ~= -240

    However, if they are evenly spaced and it didn't cross the antimeridian:
    x0 - L = 120
    x1 - x0 = 120
    R - x1 = 120

    From this, we have a delta that is guaranteed to be significantly
    large enough to tell the difference reguarless of the direction
    the antimeridian was crossed.

    However, even though the spacing was even in the source projection, it isn't
    guaranteed in the target geographic projection. So, instead of 240, 200 is used
    as it significantly larger than 120 to be sure that the antimeridian was crossed
    but smalller than 240 to account for possible irregularities in distances
    when re-projecting. Also, 200 ensures latitudes are ignored for axis order handling.
    """
    cdef Py_ssize_t arr_len = len(data)
    cdef int iii = 0
    cdef int prev_iii = arr_len - 1
    cdef double positive_min = HUGE_VAL
    cdef double min_value = HUGE_VAL
    cdef double delta = 0
    cdef int crossed_meridian_count = 0
    cdef bint positive_meridian = False

    for iii in range(0, arr_len):
        prev_iii = iii - 1
        if prev_iii == -1:
            prev_iii = arr_len - 1
        # check if crossed meridian
        delta = data[prev_iii] - data[iii]
        # 180 -> -180
        if delta >= 200:
            if crossed_meridian_count == 0:
                positive_min = min_value
            crossed_meridian_count += 1
            positive_meridian = False
        # -180 -> 180
        elif delta <= -200:
            if crossed_meridian_count == 0:
                positive_min = data[iii]
            crossed_meridian_count += 1
            positive_meridian = True
        # positive meridian side min
        if positive_meridian and data[iii] < positive_min:
            positive_min = data[iii]
        # track general min value
        if data[iii] < min_value:
            min_value = data[iii]
    if crossed_meridian_count == 2:
        return positive_min
    elif crossed_meridian_count == 4:
        # bounds extends beyond -180/180
        return -180
    return min_value


@cython.boundscheck(False)
@cython.wraparound(False)
cdef double antimeridian_max(data):
    """
    Handles the case when longitude values cross the antimeridian
    when calculating the minimum.

    Note: The data array must be in a linear ring.

    Note: This requires a densified ring with at least 2 additional
          points per edge to correctly handle global extents.

    See antimeridian_min docstring for reasoning.
    """
    cdef Py_ssize_t arr_len = len(data)
    cdef int iii = 0
    cdef int prev_iii = arr_len - 1
    cdef double negative_max = -HUGE_VAL
    cdef double max_value = -HUGE_VAL
    cdef double delta = 0
    cdef bint negative_meridian = False
    cdef int crossed_meridian_count = 0
    for iii in range(0, arr_len):
        prev_iii = iii - 1
        if prev_iii == -1:
            prev_iii = arr_len - 1
        # check if crossed meridian
        delta = data[prev_iii] - data[iii]
        # 180 -> -180
        if delta >= 200:
            if crossed_meridian_count == 0:
                negative_max = data[iii]
            crossed_meridian_count += 1
            negative_meridian = True
        # -180 -> 180
        elif delta <= -200:
            if crossed_meridian_count == 0:
                negative_max = max_value
            negative_meridian = False
            crossed_meridian_count += 1
        # negative meridian side max
        if (negative_meridian
            and (data[iii] > negative_max or negative_max == HUGE_VAL)
            and data[iii] != HUGE_VAL
        ):
            negative_max = data[iii]
        # track general max value
        if (data[iii] > max_value or max_value == HUGE_VAL) and data[iii] != HUGE_VAL:
            max_value = data[iii]
    if crossed_meridian_count == 2:
        return negative_max
    elif crossed_meridian_count == 4:
        # bounds extends beyond -180/180
        return 180
    return max_value



@cython.boundscheck(False)
@cython.wraparound(False)
def _transform_bounds(
    CRS src_crs,
    CRS dst_crs,
    double left,
    double bottom,
    double right,
    double top,
    int densify_pts,
):
    cdef double out_left = np.inf
    cdef double out_bottom = np.inf
    cdef double out_right = np.inf
    cdef double out_top = np.inf
    cdef OGRCoordinateTransformationH transform = NULL
    transform = OCTNewCoordinateTransformation(src_crs._osr, dst_crs._osr)
    transform = exc_wrap_pointer(transform)

    # OCTTransformBounds() returns TRUE/FALSE contrary to most GDAL API functions
    cdef int status = 0

    try:
        status = OCTTransformBounds(
            transform,
            left, bottom, right, top,
            &out_left, &out_bottom, &out_right, &out_top,
            densify_pts
        )
        exc_wrap_int(status == 0)
    finally:
        OCTDestroyCoordinateTransformation(transform)
    return out_left, out_bottom, out_right, out_top



"""The rio merge CLI command."""

import warnings

import click

from rasterio.enums import Resampling
from rasterio.errors import RasterioDeprecationWarning
from rasterio.rio import options
from rasterio.rio.helpers import resolve_inout
from rasterio.merge import MERGE_METHODS


def deprecated_precision(ctx, param, value):
    if value is not None:
        warnings.warn(
            "The --precision option is unused, deprecated, and will be removed in 2.0.0.",
            RasterioDeprecationWarning,
        )
    return None


@click.command(short_help="Merge a stack of raster datasets.")
@options.files_inout_arg
@options.output_opt
@options.format_opt
@options.bounds_opt
@options.resolution_opt
@click.option('--resampling',
              type=click.Choice([r.name for r in Resampling if r.value <= 7]),
              default='nearest', help="Resampling method.",
              show_default=True)
@click.option('--method',
              type=click.Choice(list(MERGE_METHODS)),
              default='first', help="Merging strategy.",
              show_default=True)
@options.nodata_opt
@options.dtype_opt
@options.bidx_mult_opt
@options.overwrite_opt
@click.option(
    "--precision",
    type=int,
    default=None,
    callback=deprecated_precision,
    help="Unused, deprecated, and will be removed in 2.0.0.",
)
@click.option(
    "--target-aligned-pixels/--no-target-aligned-pixels",
    default=False,
    help="Align the output bounds based on the resolution.",
)
@click.option(
    "--mem-limit",
    type=int,
    default=64,
    help="Limit on memory used to perform calculations, in MB.",
)
@click.option(
    "--use-highest-res/--use-first-res",
    default=False,
    help="Use the highest resolution of sources or the resolution of the first source argument.",
)
@options.creation_options
@click.pass_context
def merge(
    ctx,
    files,
    output,
    driver,
    bounds,
    res,
    resampling,
    method,
    nodata,
    dtype,
    bidx,
    overwrite,
    precision,
    target_aligned_pixels,
    mem_limit,
    use_highest_res,
    creation_options,
):
    """Copy valid pixels from input files to an output file.

    All files must have the same number of bands, data type, and
    coordinate reference system. Rotated rasters cannot be merged.

    Input files are merged in their listed order using the reverse
    painter's algorithm. If the output file exists, its values will be
    overwritten by input values.

    Geospatial bounds and resolution of a new output file in the
    units of the input file coordinate reference system may be provided
    and are otherwise taken from the first input file.

    Note: --res changed from 2 parameters in 0.25.

    \b
      --res 0.1 0.1  => --res 0.1 (square)
      --res 0.1 0.2  => --res 0.1 --res 0.2  (rectangular)

    """
    from rasterio.merge import merge as merge_tool

    output, files = resolve_inout(
        files=files, output=output, overwrite=overwrite)

    resampling = Resampling[resampling]
    if driver:
        creation_options.update(driver=driver)

    with ctx.obj["env"]:
        merge_tool(
            files,
            bounds=bounds,
            res=res,
            nodata=nodata,
            dtype=dtype,
            indexes=(bidx or None),
            resampling=resampling,
            method=method,
            target_aligned_pixels=target_aligned_pixels,
            mem_limit=mem_limit,
            use_highest_res=use_highest_res,
            dst_path=output,
            dst_kwds=creation_options,
        )



from rasterio.rio.main import main_group


def test_env_gdal_data(runner):
    """GDAL data directory printed without error."""
    result = runner.invoke(
        main_group,
        ["env", "--gdal-data"],
    )


def test_env_proj_data(runner):
    """PROJ data directory printed without error."""
    result = runner.invoke(
        main_group,
        ["env", "--proj-data"],
    )


def test_env_credentials(runner):
    """Credentials printed without error."""
    result = runner.invoke(
        main_group,
        ["env", "--credentials"],
    )


def test_env_formats(runner):
    """GTIFF format driver is registered."""
    result = runner.invoke(
        main_group,
        ["env", "--formats"],
    )
    assert "GTiff" in result.output



"""Colormap tests."""

import rasterio
from rasterio.enums import ColorInterp


def test_write_colormap(tmp_path):
    with rasterio.open("tests/data/shade.tif") as src:
        shade = src.read(1)
        profile = src.profile

    profile["driver"] = "PNG"

    with rasterio.open(tmp_path / "test.tif", "w", **profile) as dst:
        dst.write(shade, indexes=1)
        dst.write_colormap(1, {0: (255, 0, 0, 255), 255: (0, 0, 0, 0)})
        assert dst.colorinterp == (ColorInterp.palette,)
        cmap = dst.colormap(1)
        assert cmap[0] == (255, 0, 0, 255)
        assert cmap[255] == (0, 0, 0, 0)

    with rasterio.open(tmp_path / "test.tif") as src:
        assert src.colorinterp == (ColorInterp.palette,)
        cmap = src.colormap(1)
        assert cmap[0] == (255, 0, 0, 255)
        assert cmap[255] == (0, 0, 0, 0)



import pytest

from rasterio._version import get_gdal_version_info, get_geos_version, get_proj_version


@pytest.mark.parametrize(
    "key",
    [
        "--version",
        "doesnotexist",
        "VERSION_NUM",
    ],
)
def test_get_gdal_version_info(key):
    assert isinstance(get_gdal_version_info(key), str)


def test_get_proj_version():
    proj_version = get_proj_version()
    assert isinstance(proj_version, tuple)
    assert len(proj_version) == 3
    for version in proj_version:
        assert isinstance(version, int)


def test_get_geos_version():
    geos_version = get_geos_version()
    assert isinstance(geos_version, tuple)
    assert len(geos_version) == 3
    for version in geos_version:
        assert isinstance(version, int)



"""Snuggs are s-expressions for Numpy."""

# This file is a modified version of snuggs 1.4.7. The numpy
# requirement has been removed and support for keyword arguments in
# expressions has been added.
#
# Copyright 2022 Planet Labs PBC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The original license follows.
#
# Copyright (c) 2014 Mapbox
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import OrderedDict
import functools
import operator
import re
from typing import Mapping

from pyparsing import (  # type: ignore
    Keyword,
    Word,
    oneOf,
    Literal,
    QuotedString,
    ParseException,
    Forward,
    Group,
    OneOrMore,
    ParseResults,
    Regex,
    ZeroOrMore,
    alphanums,
    pyparsing_common,
    replace_with,
)

__all__ = ["eval"]
__version__ = "1.4.7"


class Context(object):
    def __init__(self):
        self._data = OrderedDict()

    def add(self, name, val):
        self._data[name] = val

    def get(self, name):
        return self._data[name]

    def lookup(self, index, subindex=None):
        s = list(self._data.values())[int(index) - 1]
        if subindex:
            return s[int(subindex) - 1]
        else:
            return s

    def clear(self):
        self._data = OrderedDict()


_ctx = Context()


class ctx(object):
    def __init__(self, kwd_dict=None, **kwds):
        self.kwds = kwd_dict or kwds

    def __enter__(self):
        _ctx.clear()
        for k, v in self.kwds.items():
            _ctx.add(k, v)
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.kwds = None
        _ctx.clear()


class ExpressionError(SyntaxError):
    """A Snuggs-specific syntax error."""

    filename = "<string>"
    lineno = 1


op_map = {
    "*": lambda *args: functools.reduce(lambda x, y: operator.mul(x, y), args),
    "+": lambda *args: functools.reduce(lambda x, y: operator.add(x, y), args),
    "/": lambda *args: functools.reduce(lambda x, y: operator.truediv(x, y), args),
    "-": lambda *args: functools.reduce(lambda x, y: operator.sub(x, y), args),
    "&": lambda *args: functools.reduce(lambda x, y: operator.and_(x, y), args),
    "|": lambda *args: functools.reduce(lambda x, y: operator.or_(x, y), args),
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
    "truth": operator.truth,
    "is": operator.is_,
    "not": operator.not_,
}

func_map: Mapping = {}
higher_func_map: Mapping = {
    "map": map,
    "partial": functools.partial,
    "reduce": functools.reduce,
    "attrgetter": operator.attrgetter,
    "methodcaller": operator.methodcaller,
    "itemgetter": operator.itemgetter,
}

nil = Keyword("null").set_parse_action(replace_with(None))
true = Keyword("true").set_parse_action(replace_with(True))
false = Keyword("false").set_parse_action(replace_with(False))


def resolve_var(source, loc, toks):
    try:
        return _ctx.get(toks[0])
    except KeyError:
        err = ExpressionError("name '{}' is not defined".format(toks[0]))
        err.text = source
        err.offset = loc + 1
        raise err


var = pyparsing_common.identifier.set_parse_action(resolve_var)
string = QuotedString("'") | QuotedString('"')
lparen = Literal("(").suppress()
rparen = Literal(")").suppress()
op = oneOf(" ".join(op_map.keys())).set_parse_action(
    lambda source, loc, toks: op_map[toks[0]]
)


def resolve_func(source, loc, toks):
    try:
        return func_map[toks[0]]
    except (AttributeError, KeyError):
        err = ExpressionError("'{}' is not a function or operator".format(toks[0]))
        err.text = source
        err.offset = loc + 1
        raise err


func = Word(alphanums + "_").set_parse_action(resolve_func)
higher_func = oneOf(" ".join(higher_func_map.keys())).set_parse_action(
    lambda source, loc, toks: higher_func_map[toks[0]]
)

func_expr = Forward()
higher_func_expr = Forward()
expr = higher_func_expr | func_expr


class KeywordArg:
    def __init__(self, name):
        self.name = name


kwarg = Regex(r":[{}]+".format(alphanums + "_")).set_parse_action(
    lambda source, loc, toks: KeywordArg(toks[0][1:])
)

operand = (
    higher_func_expr
    | func_expr
    | true
    | false
    | nil
    | var
    | kwarg
    | pyparsing_common.sci_real
    | pyparsing_common.real
    | pyparsing_common.signed_integer
    | string
)

func_expr << Group(
    lparen + (higher_func_expr | op | func) + OneOrMore(operand) + rparen
)

higher_func_expr << Group(
    lparen
    + higher_func
    + (nil | higher_func_expr | op | func | OneOrMore(operand))
    + ZeroOrMore(operand)
    + rparen
)


def processArg(arg):
    if isinstance(arg, ParseResults):
        return processList(arg)
    else:
        return arg


def processList(lst):
    items = [processArg(x) for x in lst[1:]]
    args = []
    kwds = {}

    # An iterator is used instead of implicit iteration to allow
    # skipping ahead in the keyword argument case.
    itemitr = iter(items)

    for item in itemitr:
        if isinstance(item, KeywordArg):
            # The next item after the keyword arg marker is its value.
            # This advances the iterator in a way that is compatible
            # with the for loop.
            val = next(itemitr)
            key = item.name
            kwds[key] = val
        else:
            args.append(item)

    func = processArg(lst[0])

    # list and tuple are two builtins that take a single argument,
    # whereas args is a list. On a KeyError, the call is retried
    # without arg unpacking.
    try:
        return func(*args, **kwds)
    except TypeError:
        return func(args, **kwds)


def handleLine(line):
    try:
        result = expr.parseString(line)
        return processList(result[0])
    except ParseException as exc:
        text = str(exc)
        m = re.search(r"(Expected .+) \(at char (\d+)\), \(line:(\d+)", text)
        msg = m.group(1)
        if "map|partial" in msg:
            msg = "expected a function or operator"
        err = ExpressionError(msg)
        err.text = line
        err.offset = int(m.group(2)) + 1
        raise err


def eval(source, kwd_dict=None, **kwds):
    """Evaluate a snuggs expression.

    Parameters
    ----------
    source : str
        Expression source.
    kwd_dict : dict
        A dict of items that form the evaluation context. Deprecated.
    kwds : dict
        A dict of items that form the valuation context.

    Returns
    -------
    object

    """
    kwd_dict = kwd_dict or kwds
    with ctx(kwd_dict):
        return handleLine(source)
